import logging
import os

import Bio
import fstlib
import matplotlib
import numpy as np
import pandas as pd

from medicc import plot, tools

matplotlib.use("Agg")
logger = logging.getLogger(__name__)


def read_and_parse_input_data(filename, normal_name='diploid', input_type='tsv', separator='X', 
                              allele_columns=['cn_a', 'cn_b'], maxcn=8, total_copy_numbers=False):
    ## Read in input data
    if input_type.lower() == "fasta" or input_type.lower() == 'f':
        logger.info("Reading FASTA input.")
        input_df = _read_fasta_as_dataframe(filename, separator=separator, allele_columns=allele_columns, maxcn=maxcn)
    elif input_type.lower() == "tsv" or input_type.lower() == 't':
        logger.info("Reading Refphase TSV input.")
        input_df = _read_tsv_as_dataframe(filename, allele_columns=allele_columns, maxcn=maxcn)
    else:
        raise MEDICCIOError("Unknown input type, possible options are 'fasta' or 'tsv'.")

    if len(allele_columns) == 1 and not total_copy_numbers:
        logger.warn('You have provided only one allele column but the --total-copy-numbers flag was not set')
    if total_copy_numbers and not len(allele_columns) == 1:
        raise MEDICCIOError("You have set the --total-copy-numbers flag but provided more than one allele column. "
                            "Set allele columns with the flag --input-allele-columns")

    input_df.columns.name = 'allele'
    input_df_stacked = input_df.stack('allele').unstack('sample_id').T

    duplicated_entries = input_df_stacked.duplicated(keep=False)
    if duplicated_entries.any():
        logger.warn("Duplicated entries found in input data: "
                    f"{input_df_stacked.index[duplicated_entries]}")

    normal_value = '2' if total_copy_numbers else '1'
    normal_samples = np.setdiff1d(input_df_stacked.index[
        (input_df_stacked == normal_value).all(axis=1)], normal_name)
    if len(normal_samples) > 0:
        logger.warn(f"Normal samples found in input data: {normal_samples}")
    
    ## Add normal sample if needed
    input_df = add_normal_sample(input_df, normal_name, allele_columns=allele_columns, 
                                    total_copy_numbers=total_copy_numbers)
    nsamples = input_df.index.get_level_values('sample_id').unique().shape[0]
    nchr = input_df.index.get_level_values('chrom').unique().shape[0]
    nsegs = input_df.loc[normal_name,:].shape[0]
    logger.info("Read %d samples, %d chromosomes, %d segments per sample", nsamples, nchr, nsegs)
    return input_df


def read_fst(user_fst=None, no_wgd=False, n_wgd=None, total_copy_numbers=False, wgd_x2=False):
    """Simple wrapper for loading the FST using the fstlib read function. """

    objects_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "objects")

    if user_fst is not None:
        fst_path = user_fst
    elif no_wgd:
        fst_path = os.path.join(objects_dir, 'no_wgd_asymm.fst')
    elif wgd_x2:
        fst_path = os.path.join(objects_dir, 'wgd_x2_asymm.fst')
        if n_wgd == 1:
            fst_path = os.path.join(objects_dir, 'wgd_x2_1_asymm.fst')
    elif total_copy_numbers:
        fst_path = os.path.join(objects_dir, 'wgd_total_cn_asymm.fst')
    elif n_wgd is not None and int(n_wgd) <= 3:
        fst_path = os.path.join(objects_dir, 'wgd_{}_asymm.fst'.format(int(n_wgd)))
    else:
        fst_path = os.path.join(objects_dir, 'wgd_asymm.fst')

    return fstlib.read(fst_path)


def load_main_fsts(return_symbol_table=False):
    asymm_fst = read_fst()
    asymm_fst_nowgd = read_fst(no_wgd=True)
    asymm_fst_1_wgd = read_fst(n_wgd=1)
    asymm_fst_2_wgd = read_fst(n_wgd=2)

    if return_symbol_table:
        symbol_table = asymm_fst.input_symbols()
        return asymm_fst, asymm_fst_nowgd, asymm_fst_1_wgd, asymm_fst_2_wgd, symbol_table
    else:
        return asymm_fst, asymm_fst_nowgd, asymm_fst_1_wgd, asymm_fst_2_wgd


def validate_input(input_df, symbol_table):
    # Check the number of alleles
    if len(input_df.columns)>2:
        raise MEDICCIOError("More than 2 alleles are currently not supported.")

    if len(input_df.columns)==0:
        raise MEDICCIOError("No alleles found.")

    # Check if index of dataframe is sorted
    if not input_df.index.is_monotonic_increasing:
        raise MEDICCIOError("DataFrame index must be sorted.")

    # Check if all samples have same number of segments
    if input_df.unstack('sample_id').isna().sum().sum() != 0:
        raise MEDICCIOError("The samples have different number of segments!\n"
                            "Total number of unique segments: {}\n".format(len(input_df.unstack('sample_id'))))

    # Check if symbols are in symbol table
    alphabet = {x[1] for x in symbol_table}
    data_chars = set(input_df.values.flatten())
    if not data_chars.issubset(alphabet):
        not_in_set = data_chars.difference(alphabet)
        raise MEDICCIOError("Not all input symbols are contained in symbol table. Offending symbols: %s" % str(not_in_set))

    # Check data type start and end columns
    if (input_df.index.get_level_values('start').dtype != np.int or 
        input_df.index.get_level_values('end').dtype != np.int):
        raise MEDICCIOError("Start and end columns must be of type: integer.")

    # Check data type payload columns - these should all be of type str (object)
    if not np.all([pd.api.types.is_string_dtype(x) for x in input_df.dtypes]):
        raise MEDICCIOError("Payload columns must be of type: string.")

    # Check if index of dataframe is sorted
    if not input_df.index.is_monotonic_increasing:
        raise MEDICCIOError("DataFrame index must be sorted.")

    logger.info('Input data is valid!')

def filter_by_segment_length(input_df, filter_size):
    segment_length = input_df.eval('end-start')
    return input_df.loc[segment_length > float(filter_size)]

def _read_tsv_as_dataframe(path, allele_columns=['cn_a','cn_b'], maxcn=8):
    logger.info("Reading TSV file %s", path)
    input_file = pd.read_csv(path, sep = "\t")
    columnn_names = ['sample_id', 'chrom', 'start', 'end'] + allele_columns
    if len(np.setdiff1d(columnn_names, input_file.columns)) > 0:
        raise MEDICCIOError(f"TSV file needs the following columns: sample_id, chrom, start, end and the allele columns ({allele_columns})"
                            "\nMissing columns are: {}".format(
                                np.setdiff1d(columnn_names, input_file.columns)))

    logger.info("Successfully read input file. Using columns {%s}" % ', '.join(columnn_names))
    input_file = input_file[columnn_names]
    for c in allele_columns:
        if input_file[c].dtype in [np.dtype('float64'), np.dtype('float32')]:
            logger.warning("Floating point payload! I will round, but this might not be intended.")
            input_file[c] = input_file[c].round().astype('int')
        if input_file[c].dtype in [np.dtype('int64'), np.dtype('int32')]:
            if np.any(input_file[c]>maxcn):
                logger.warning("Integer CN > maxcn %d, capping.", maxcn)
                input_file[c] = np.fmin(input_file[c], maxcn)
    input_file['chrom'] = tools.format_chromosomes(input_file['chrom'])
    if input_file['chrom'].apply(lambda x: "Y" in str(x)).any():
        logger.warn("Y chromosome detected in input. This might cause errors down the line!")
    input_file.set_index(['sample_id', 'chrom', 'start', 'end'], inplace=True)
    input_file.sort_index(inplace=True)
    input_file[allele_columns] = input_file[allele_columns].astype(str)

    return input_file

def _read_fasta_as_dataframe(infile: str, separator: str = 'X', allele_columns = ['cn_a','cn_b'], maxcn: int = 8):
    """Reads FASTA decriptor file (old MEDICC input format) and reads the corresponding FASTA files to generate
    a data frame with the same format as the refphase input (TSV) format. """
    logger.info("Reading FASTA dataset from description file %s.", infile)
    description_file = pd.read_csv(infile,
                                    delim_whitespace = True,
                                    header = None,
                                    names = ["chrom",] + allele_columns,
                                    usecols = ["chrom",] + allele_columns)
    description_file.set_index('chrom', inplace=True)
    description_file.columns.name='allele'
    description_file = description_file.stack()
    inpath = os.path.dirname(infile)

    payload = []
    for filename in description_file:
        with open(os.path.join(inpath, filename), 'r') as fd:
            text = fd.read()
            df = pd.DataFrame([s.strip().split('\n') for s in text.split('>') if s.strip() !=''], columns=['sample_id', 'cnp'])
            df.set_index('sample_id', inplace=True)
            payload.append(df)

    payload = pd.concat(payload, keys=description_file.index, names=description_file.index.names)

    ## deal with the case where multiple chromosomes are given in one string
    payload=payload.cnp.str.split(separator, expand=True)
    if payload.shape[1]>1: ## multiple columns
        payload.columns.name='id'
        payload=payload.unstack('chrom').reorder_levels(['chrom','id'],axis=1)
        payload.columns = ["%s%d" % (c,i+1) for c,i in payload.columns]
        payload.columns.name = 'chrom'
        payload = payload.stack()
        payload = payload.reorder_levels([1,2,0]).sort_index()
    else:
        payload = payload.iloc[:,0]
        payload = payload.reorder_levels([2,0,1]).sort_index()

    payexpand = [pd.DataFrame(list(s), columns=['cn']) for s in payload]
    paylong = pd.concat(payexpand, keys=payload.index, names=payload.index.names + ['segment'])
    paylong['cn'] = paylong['cn'].apply(tools.hex2int)
    if np.any(paylong['cn'] > maxcn):
        logger.warning("Integer CN > maxcn %d, capping.", maxcn)
        paylong['cn'] = np.fmin(paylong['cn'], maxcn)
    paylong['cn'] = paylong['cn'].astype(str)
    result = paylong.unstack(['allele'])
    result = result.droplevel(0, axis=1).reset_index()
    result.columns.name = None
    result['start'] = result['segment']
    result['end'] = result['start']+1
    result = result[['sample_id','chrom','start','end'] + allele_columns]
    result['chrom'] = tools.format_chromosomes(result['chrom'])
    result.sort_values(['sample_id', 'chrom', 'start', 'end'], inplace=True)
    result.set_index(['sample_id', 'chrom', 'start', 'end'], inplace=True)
    result.sort_index(inplace=True)

    return result

def add_normal_sample(df, normal_name, allele_columns=['cn_a','cn_b'], total_copy_numbers=False):
    """Adds an artificial normal samples with the supplied name to the data frame.
    The normal sample has CN=1 on all supplied alleles. """
    samples = df.index.get_level_values('sample_id').unique()

    if total_copy_numbers:
        normal_value = '2'
    else:
        normal_value = '1'

    if normal_name is not None and normal_name not in samples:
        logger.info("Normal sample '%s' not found, adding artifical normal by the name: '%s'.", normal_name, normal_name)
        tmp = df.unstack('sample_id')
        for col in allele_columns:
            tmp.loc[:, (col, normal_name)] = normal_value
        tmp = tmp.stack('sample_id')
        tmp = tmp.reorder_levels(['sample_id', 'chrom', 'start', 'end']).sort_index()
    else:
        logger.info("Sample '%s' was found in data is is used as normal", normal_name)
        if np.any(df.loc[normal_name] == '0'):
            logger.warn("The provided normal sample contains segments with copy number 0. "
                        "If any other sample has non-zero values in these segments, MEDICC will crash")
        if np.any(df.loc[normal_name] != normal_value):
            logger.warn("The provided normal sample contains segments with copy number != {}.".format(normal_value))

        tmp = df

    return tmp


def write_tree_files(tree, out_name: str, plot_tree=True, draw_ascii=False, normal_name='diploid'):
    """Writes a Newick, PhyloXML, Ascii graphic and PNG grahic file of the tree. """
    Bio.Phylo.write(tree, out_name + ".new", "newick")
    Bio.Phylo.write(tree, out_name + ".xml", "phyloxml")

    if draw_ascii:
        with open(out_name + ".txt", "w") as f:
            Bio.Phylo.draw_ascii(tree, file = f)

    if plot_tree:
        plot.plot_tree(tree,
                       output_name=out_name,
                       normal_name=normal_name,
                       label_func=lambda x: x if 'internal' not in x else '',
                       show_branch_lengths=True)


def write_branch_lengths(tree, out_name: str):
    """Writes a file with the branch lengths of the tree."""
    with open(out_name, "w") as f:
        for node in tree.find_clades():
            if node.name is not None and node.name != 'diploid':
                f.write("{}\t{}\n".format(node.name, node.branch_length))


def write_pairwise_distances(sample_labels, pairwise_distances, filename_prefix):
    """Write the pairwise distance matrix as a tsv."""

    if not isinstance(pairwise_distances, pd.DataFrame):
        pairwise_distances = pd.DataFrame(pairwise_distances, columns=sample_labels, index=sample_labels)
    pairwise_distances.to_csv(f"{filename_prefix}.tsv", sep='\t')


def import_tree(tree_file, normal_name='diploid', file_format='newick'):
    """Loads a phylogenetic tree in the given format and roots it at the normal sample. """
    tree = Bio.Phylo.read(tree_file, file_format)
    input_tree = Bio.Phylo.BaseTree.copy.deepcopy(tree)
    tmpsearch = [c for c in input_tree.find_clades(name = normal_name)]
    normal_name = tmpsearch[0]
    root_path = input_tree.get_path(normal_name)[::-1]

    if len(root_path) > 1:
        new_root = root_path[1]
        input_tree.root_with_outgroup(new_root)
    else:
        pass

    # check that internal node names are unique
    node_names = [c.name for c in input_tree.find_clades() if c.name is not None]
    if len(node_names) != len(np.unique(node_names)):
        logger.warning("Internal node names are not unique. This will cause problems in MEDICC2.")

    return input_tree


def read_bed_file(filename):

    data = pd.read_csv(filename, header=None, comment='#', sep='\t')

    if len(data.columns) != 4:
        print('WARNING: expected 4 columns')
        return None

    data.columns = ['Chromosome', 'Start', 'End', 'name']
    return data


class MEDICCIOError(Exception):
    pass



import copy
import logging
import os

import numpy as np
import pandas as pd
from Bio.Phylo.Consensus import _BitString, get_support
from joblib.parallel import Parallel, delayed

from medicc import io, tools
from medicc.core import main

# tqdm can be used for progress bars
try: 
    from tqdm.auto import tqdm
except ImportError:
    def tqdm(x, disable):
        return x

logger = logging.getLogger(__name__)

def chr_wise_bootstrap_df(input_df):
    """Creates a bootstrap dataframe based on the original data. From the N original chromosomes in
    the dataset we draw N chromosomes *with* replacement to create a new bootstrap dataframe.
    """

    chroms = input_df.index.get_level_values('chrom').unique()
    cur_chroms = np.random.choice(chroms, len(chroms), replace=True)

    bootstrap_df = pd.DataFrame(columns=input_df.reset_index().columns)

    for i, cur_chrom in enumerate(cur_chroms):
        cur_data = input_df.reset_index().loc[(input_df.reset_index()['chrom'] == cur_chrom).values]
        cur_data['chrom'] = 'chr{}'.format(i+1)
        bootstrap_df = pd.concat([bootstrap_df, cur_data])

    bootstrap_df['chrom'] = tools.format_chromosomes(bootstrap_df['chrom'])
    bootstrap_df.set_index(['sample_id', 'chrom', 'start', 'end'], inplace=True)
    bootstrap_df.sort_index(inplace=True)

    logger.debug("Created chr-wise bootstrap dataframe")

    return bootstrap_df


def segment_wise_jacknife_df(input_df):
    """Creates a jackknife dataframe based on the original data. From the N original segments in
    the dataset we draw N segments *with* replacement and keeping their original order. For 
    duplicate segments only one copy will be kept. On average this will result in (1-1/e) * N segments. 
    """
    jacknife_df = input_df.copy()
    jacknife_df['segment'] = jacknife_df.reset_index()[['chrom', 'start']].astype(
        str).apply(lambda x: '-'.join(x), axis=1).values
    segments = jacknife_df['segment'].unique()

    cur_segments = np.unique(np.random.choice(segments, size=len(segments), replace=True))
    jacknife_df = jacknife_df.loc[jacknife_df['segment'].map(lambda x: x in cur_segments)].drop('segment', axis=1)

    logger.debug(
        "Created segment-wise jackknife dataframe with {} out of the {} segments".format(len(cur_segments), 
                                                                                         len(segments)))

    return jacknife_df


def bootstrap_shuffle_chroms(input_df, normal_name='diploid'):
    """Creates a bootstrap dataframe based on the original data. The chromosome-wise copy-number data 
    will be shuffled between samples. The corresponding dataframe will be generally similar to the
    original data but have to no the distinct features of the original data.
    It can therefore not be used to calculate branch support or similar scores.
    """

    chroms = list(input_df.index.get_level_values('chrom').unique())
    sample_labels = [s for s in input_df.reset_index()['sample_id'].unique() if s != normal_name]
    shuffled_samples = copy.copy(sample_labels)
    cur_df = copy.copy(input_df).reset_index()

    for chrom in chroms:
        np.random.shuffle(shuffled_samples)
        renaming = {label: shuffled_label for label,
                    shuffled_label in zip(sample_labels, shuffled_samples)}
        renaming.update({normal_name: normal_name})
        cur_df.loc[cur_df['chrom'] == chrom, 'sample_id'] = cur_df.loc[cur_df['chrom']
                                                                       == chrom, 'sample_id'].map(lambda x: renaming[x])

    cur_df.sort_values(['sample_id', 'chrom', 'start', 'end'], inplace=True)
    cur_df.set_index(['sample_id', 'chrom', 'start', 'end'], inplace=True)
    cur_df.sort_index(inplace=True)

    return cur_df


def _bitstrs(tree):
    """Create a signature for the tree topology. Each internal node has a binary string with length
    equal to the number of leaf nodes with values 1 if the internal node is downstream of a particular
    leaf node and 0 otherwise.
    The final list of strings is a unique identifier of the tree topology.
    """
    # from https://biopython.org/wiki/Phylo
    bitstrs = set()
    term_names = [term.name for term in tree.get_terminals()]
    term_names.sort()
    for clade in tree.get_nonterminals():
        clade_term_names = [term.name for term in clade.get_terminals()]
        boolvals = [name in clade_term_names for name in term_names]
        bitstr = _BitString("".join(map(str, map(int, boolvals))))
        bitstrs.add(bitstr)
    return bitstrs


def compare_trees(tree1, tree2, fail_on_different_terminals=True):
    """Compare to trees topologies. The length of the branches are not regarded.

    If fail_on_different_terminals is set to True the function fails if the terminals are not the
    same. This prevents accidental comparison of the wrong trees.
    
    Modified from https://biopython.org/wiki/Phylo
    """
    term_names1 = [term.name for term in tree1.get_terminals()]
    term_names2 = [term.name for term in tree2.get_terminals()]
    # false if terminals are not the same
    if set(term_names1) != set(term_names2):
        if fail_on_different_terminals:
            raise ValueError("Leaf names of the two trees are not the same:\ntree 1: {}\ntree 2:{}".format(
                set(term_names1), set(term_names2)))
        else:
            return False
    # true if _BitStrings are the same
    return _bitstrs(tree1) == _bitstrs(tree2)


def _single_bootstrap_run(input_df, fst, bootstrap_method, i, N_bootstrap, normal_name='diploid'):
    cur_df = bootstrap_method(input_df)
    _, _, _, cur_final_tree, _, _ = main(
        cur_df,
        fst,
        normal_name,
        input_tree=None,
        ancestral_reconstruction=False,
        chr_separator='X')

    if (i+1) % (N_bootstrap//5) == 0:
        logger.info('{}/{} ({}%) bootstrap runs completed'.format(i + 1, N_bootstrap, 
                                                                  int((i+1)/N_bootstrap*100)))

    return cur_final_tree


def run_bootstrap(input_df,
                  original_tree=None,
                  N_bootstrap=50,
                  method='chr-wise',
                  wgd=True,
                  seed=None,
                  normal_name='diploid',
                  show_progress=True,
                  n_cores=None):
    """Run a given number of bootstrapping steps on the original data. 

    From the original data either a set of chromosome-wise bootstrap or segment-wise jackknife datasets
    are created. For every one of these datasets the optimal MEDICC tree is reconstructed and added
    to a tree dataframe. If the original tree is provided support values for the individual branches
    are calculated.
    """

    # tqdm can be used for progress bars
    try:
        from tqdm.auto import tqdm
    except ImportError:
        def tqdm(x, disable):
            return x

    if not show_progress:
        def tqdm(x, disable):
            return x

    if method == 'chr-wise':
        bootstrap_method = chr_wise_bootstrap_df
        logger.info('Starting {} chr-wise bootstrap runs'.format(N_bootstrap))
    elif method == 'segment-wise':
        logger.info('Starting {} segment-wise jackknife runs'.format(N_bootstrap))
        bootstrap_method = segment_wise_jacknife_df
    else:
        raise ValueError('method has to be either chr-wise or segment-wise')

    if seed is not None:
        np.random.seed(seed)

    if wgd:
        fst = io.read_fst()
    else:
        fst = io.read_fst(no_wgd=True)

    if original_tree is not None:
        trees = {original_tree: [0, 'original']}
    else:
        trees = dict()

    # Run the actual bootstrapping steps
    if n_cores is not None and n_cores > 1:
        initial_trees = Parallel(n_jobs=n_cores)(delayed(_single_bootstrap_run)(
            input_df, fst, bootstrap_method, i, N_bootstrap, normal_name)
            for i in range(N_bootstrap))
    else:
        initial_trees = []
        for i in tqdm(range(N_bootstrap), disable=not show_progress):
            cur_tree = _single_bootstrap_run(
                input_df, fst, bootstrap_method, i, N_bootstrap, normal_name)
            initial_trees.append(cur_tree)

    # delete duplicate trees
    for new_tree in initial_trees:
        for tree in trees.keys():
            if compare_trees(new_tree, tree):
                trees[tree][0] += 1
                break
        else:
            trees[new_tree] = [1, '']

    trees_df = pd.DataFrame(trees).T.reset_index().sort_values(
        0, ascending=False).reset_index(drop=True)
    trees_df.rename({0: 'count', 1: 'note', 'index': 'tree'}, axis=1, inplace=True)
    trees_df['freq'] = trees_df['count'] / trees_df['count'].sum()
    logger.info('A total of {} bootstrap trees were created'.format(len(trees_df)))

    # Calculate the support tree from the bootstrap trees
    if original_tree is not None:
        logger.info('Input tree provided. Calculating support tree')
        trees_list = []
        for _, row in trees_df.iterrows():
            trees_list += row['count'] * [row['tree']]
        support_tree = get_support(original_tree, trees_list)

        # Remove branch support for first branch because it is meaningless for rooted trees
        for clade in support_tree.root.clades:
            clade.confidence = None

    else:
        logger.debug('No input tree provided. Support tree is not calculated')
        support_tree = None

    return trees_df, support_tree


def bootstrap_hypothesis_test(input_df, hypothesis, trees_df=None, **kwargs):
    """Test a hypothesis on a set of bootstrap trees. `hypothesis` has to be a function.
    """
    if trees_df is None:
        trees_df, _ = run_bootstrap(input_df, **kwargs)

    trees_df['hypothesis'] = trees_df['tree'].map(hypothesis)

    return trees_df, trees_df.groupby('hypothesis')['freq'].sum()



import os
import pathlib
import subprocess
import time

import numpy as np
import pandas as pd
import pytest


def test_medicc_help_box():
    "Just testing that medicc can be started"
    process = subprocess.Popen(['python', "medicc2", "--help"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    assert process.returncode == 0


def test_medicc_with_simple_example():
    "Testing small example"
    output_dir = 'examples/test_output'
    process = subprocess.Popen(['python', "medicc2", "examples/simple_example/simple_example.tsv", 
                                output_dir, "--plot", "both"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['simple_example_cn_profiles.pdf', 'simple_example_final_cn_profiles.tsv',
                      'simple_example_final_tree.new', 'simple_example_final_tree.png',
                      'simple_example_final_tree.xml', 'simple_example_pairwise_distances.tsv',
                      'simple_example_summary.tsv', 'simple_example_copynumber_events_df.tsv',
                      'simple_example_events_overlap.tsv', 'simple_example_branch_lengths.tsv',
                      'simple_example_cn_profiles_heatmap.pdf']
    all_files_exist = [os.path.isfile(os.path.join('examples/test_output/', f)) for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'simple_example')
    output_df = pd.read_csv(os.path.join(output_dir, "simple_example_final_cn_profiles.tsv"), sep='\t')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"

    assert output_df['is_gain'].sum() == 7, "Number of gains in output_df is not 7"
    assert output_df['is_loss'].sum() == 4, "Number of losses in output_df is not 4"

def test_medicc_with_testing_example():
    "Testing testing example"
    output_dir = 'examples/test_output'
    process = subprocess.Popen(['python', "medicc2", "examples/testing_example/testing_example.tsv", 
                                output_dir],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['testing_example_cn_profiles.pdf', 'testing_example_final_cn_profiles.tsv',
                      'testing_example_final_tree.new', 'testing_example_final_tree.png',
                      'testing_example_final_tree.xml', 'testing_example_pairwise_distances.tsv',
                      'testing_example_summary.tsv', 'testing_example_copynumber_events_df.tsv',
                      'testing_example_events_overlap.tsv', 'testing_example_branch_lengths.tsv']
    all_files_exist = [os.path.isfile(os.path.join('examples/test_output/', f)) for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'testing_example')
    output_df = pd.read_csv(os.path.join(output_dir, "testing_example_final_cn_profiles.tsv"), sep='\t')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"

    assert output_df['is_gain'].sum() == 192, "Number of gains in output_df is not 192"
    assert output_df['is_loss'].sum() == 173, "Number of losses in output_df is not 173"


def test_medicc_with_OV03_04():
    "Testing testing example"
    output_dir = 'examples/test_output'
    process = subprocess.Popen(['python', "medicc2", "examples/OV03-04/OV03-04_descr.txt", 
                                output_dir, "-i", "fasta", "--normal-name", "OV03-04_diploid",
                                "--plot", "both"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['OV03-04_descr_cn_profiles.pdf', 'OV03-04_descr_final_cn_profiles.tsv',
                      'OV03-04_descr_final_tree.new', 'OV03-04_descr_final_tree.png',
                      'OV03-04_descr_final_tree.xml', 'OV03-04_descr_pairwise_distances.tsv',
                      'OV03-04_descr_summary.tsv', 'OV03-04_descr_copynumber_events_df.tsv',
                      'OV03-04_descr_events_overlap.tsv', 'OV03-04_descr_branch_lengths.tsv',
                      'OV03-04_descr_cn_profiles_heatmap.pdf']
    all_files_exist = [os.path.isfile(os.path.join('examples/test_output/', f)) for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'OV03-04_descr')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"


def test_medicc_with_testing_example_total_copy_numbers():
    "Testing small example"
    output_dir = 'examples/test_output_total_cn'
    process = subprocess.Popen(['python', "medicc2", "examples/testing_example/testing_example.tsv", 
                                output_dir, "--total-copy-numbers", 
                                "--input-allele-columns", "cn_a"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['testing_example_cn_profiles.pdf', 'testing_example_final_cn_profiles.tsv',
                      'testing_example_final_tree.new', 'testing_example_final_tree.png',
                      'testing_example_final_tree.xml', 'testing_example_pairwise_distances.tsv',
                      'testing_example_summary.tsv', 'testing_example_copynumber_events_df.tsv',
                      'testing_example_events_overlap.tsv', 'testing_example_branch_lengths.tsv']
    all_files_exist = [os.path.isfile(os.path.join(output_dir, f))
                       for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'testing_example')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"


def test_medicc_with_testing_example_nowgd():
    "Testing small example"
    output_dir = 'examples/test_output_nowgd'
    process = subprocess.Popen(['python', "medicc2", "examples/testing_example/testing_example.tsv", 
                                output_dir, "--no-wgd"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['testing_example_cn_profiles.pdf', 'testing_example_final_cn_profiles.tsv',
                      'testing_example_final_tree.new', 'testing_example_final_tree.png',
                      'testing_example_final_tree.xml', 'testing_example_pairwise_distances.tsv',
                      'testing_example_summary.tsv', 'testing_example_copynumber_events_df.tsv',
                      'testing_example_events_overlap.tsv', 'testing_example_branch_lengths.tsv']
    all_files_exist = [os.path.isfile(os.path.join(output_dir, f))
                       for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'testing_example')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"


def test_medicc_with_testing_example_WGD_x2():
    "Testing small example"
    output_dir = 'examples/test_output_wgd_x2'
    process = subprocess.Popen(['python', "medicc2", "examples/testing_example/testing_example.tsv", 
                                output_dir, "--wgd-x2"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['testing_example_cn_profiles.pdf', 'testing_example_final_cn_profiles.tsv',
                      'testing_example_final_tree.new', 'testing_example_final_tree.png',
                      'testing_example_final_tree.xml', 'testing_example_pairwise_distances.tsv',
                      'testing_example_summary.tsv', 'testing_example_copynumber_events_df.tsv',
                      'testing_example_events_overlap.tsv', 'testing_example_branch_lengths.tsv']
    all_files_exist = [os.path.isfile(os.path.join(output_dir, f))
                       for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'testing_example')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"


def test_medicc_with_multiple_cores():
    "Testing small example"
    output_dir = 'examples/test_output_multiple_cores'
    process = subprocess.Popen(['python', "medicc2", "examples/simple_example/simple_example.tsv", 
                                output_dir, "-j 4"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = ['simple_example_cn_profiles.pdf', 'simple_example_final_cn_profiles.tsv',
                      'simple_example_final_tree.new', 'simple_example_final_tree.png',
                      'simple_example_final_tree.xml', 'simple_example_pairwise_distances.tsv',
                      'simple_example_summary.tsv', 'simple_example_copynumber_events_df.tsv',
                      'simple_example_events_overlap.tsv', 'simple_example_branch_lengths.tsv']
    all_files_exist = [os.path.isfile(os.path.join('examples/test_output_multiple_cores/', f))
                       for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, 'simple_example')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"


def test_medicc_with_bootstrap():
    "Testing bootstrap workflow"
    output_dir = 'examples/test_output_bootstrap'
    process = subprocess.Popen(['python', "medicc2",
                                "examples/simple_example/simple_example.tsv",
                                output_dir,
                                "--bootstrap-nr", "5"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    support_tree_exists = os.path.isfile('examples/test_output_bootstrap/simple_example_support_tree.new')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, 'Error while running MEDICC'
    assert support_tree_exists, "Support tree file was not created"


gundem_et_al_2015_patients = ['PTX004', 'PTX005', 'PTX006', 'PTX007', 'PTX008', 
                              'PTX009', 'PTX010', 'PTX011', 'PTX012', 'PTX013']
extra_condition = ['normal', 'no_wgd', 'total_cn', 'wgd_x2']
@pytest.mark.parametrize("patient", gundem_et_al_2015_patients)
@pytest.mark.parametrize("extra_condition", extra_condition)
def test_gundem_et_al_2015(patient, extra_condition):
    "Testing if running of all Gundem data works"

    output_dir = f"examples/test_output_{patient}"
    command = ['python', "medicc2", f"examples/gundem_et_al_2015/{patient}_input_df.tsv", output_dir]
    if extra_condition == 'normal':
        pass
    elif extra_condition == 'no_wgd':
        command.append('--no-wgd')
    elif extra_condition == 'total_cn':
        command += ['--total-copy-numbers', '--input-allele-columns', 'cn_a']
    elif extra_condition == 'wgd_x2':
        command.append('--wgd-x2')
        
    process = subprocess.Popen(command,
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    expected_files = [f'{patient}_input_df_cn_profiles.pdf', f'{patient}_input_df_final_cn_profiles.tsv',
                      f'{patient}_input_df_final_tree.new', f'{patient}_input_df_final_tree.png',
                      f'{patient}_input_df_final_tree.xml', f'{patient}_input_df_pairwise_distances.tsv',
                      f'{patient}_input_df_summary.tsv', f'{patient}_input_df_copynumber_events_df.tsv',
                      f'{patient}_input_df_events_overlap.tsv', f'{patient}_input_df_branch_lengths.tsv']

    all_files_exist = [os.path.isfile(os.path.join(output_dir, f)) for f in expected_files]
    nr_events, tree_size = get_number_of_events(output_dir, f'{patient}_input_df')
    subprocess.Popen(["rm", output_dir, "-rf"])

    assert process.returncode == 0, f'Error while running MEDICC for Gundem et al patient {patient}'
    assert np.all(all_files_exist), "Some files were not created! \nMissing files are: {}".format(
        np.array(expected_files)[~np.array(all_files_exist)])
    assert nr_events == tree_size, f"Number of events is {nr_events}, but tree size is {tree_size}"


all_notebooks = [x for x in os.listdir('notebooks') if '.py' in x]
@pytest.mark.parametrize("notebook", all_notebooks)
def test_all_notebooks(notebook):
    "Testing if all notebooks work"

    process = subprocess.Popen(['python', 
                                f"notebooks/{notebook}"],
                               stdout=subprocess.PIPE,
                               cwd=pathlib.Path(__file__).parent.parent.absolute())

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    assert process.returncode == 0, f'Error while running notebook {notebook}'


all_files = [x for x in os.listdir('Figures_Kaufmann_et_al_2021') if '.py' in x]
@pytest.mark.parametrize("cur_file", all_files)
def test_figures_for_publication(cur_file):
    "Testing if all Figure scripts work"
    print(os.path.join(pathlib.Path(__file__).parent.parent.absolute(),
                       "Figures_Kaufmann_et_al_2021"))

    process = subprocess.Popen(['python',
                                f"{cur_file}"],
                                stdout=subprocess.PIPE,
                                cwd=os.path.join(pathlib.Path(__file__).parent.parent.absolute(),
                                                 "Figures_Kaufmann_et_al_2021"))

    while process.poll() is None:
        # Process hasn't exited yet
        time.sleep(0.5)

    assert process.returncode == 0, f'Error while running {cur_file}'


def get_number_of_events(output_dir, file_prefix):
    with open(os.path.join(output_dir, f"{file_prefix}_copynumber_events_df.tsv"), 'r') as f:
        events = f.readlines()
    nr_events = len(events) - 1

    with open(os.path.join(output_dir, f"{file_prefix}_summary.tsv"), 'r') as f:
        summary = f.readlines()
    tree_size = float(summary[2].split('\t')[1].rstrip())

    return nr_events, tree_size



'''
Created on 09 Dec 2019

@author: mp 
'''

import logging

import Bio
import Bio.Nexus
import Bio.Phylo
import numpy as np

SS_TAXON_SEPERATOR = ","
CHR_SEPARATOR = "X"

logger = logging.getLogger(__name__)

class NeighbourJoining(object):
    """ computes a neighbour joining tree from a distance matrix """
    
    def __init__(self, distance_matrix, taxa):
        """ constructor:
        requires distance matrix as numpy array and list of taxa 
        names in the same order as in the matrix """
        
        self.distmat = distance_matrix
        self.taxa = taxa
        self.tree = None
        self.internal_nodes = []
        self.log = logger
        self._run_nj()
    
    def _run_nj(self):
        """ main algorithm routine """
        
        self.log.info("Neighbor joining:")
        name_clade_map = dict()
        internal_node_list = list()
        sequence_files = list(self.taxa) ## create mutable copy
        distances = self.distmat.copy()
        
        while len(sequence_files) > 2: ## do while still more than two files present
            q_matrix = self._nj_create_q_matrix(distances)
            (left, right) = self._nj_nearest_neighbours(q_matrix)
            left_name = sequence_files[left]
            right_name = sequence_files[right]
            self.log.debug("Choosing %s with distance %f" % ((left_name, right_name), distances[left,right]))
            new_distances = self._nj_get_distances_to_new_node((left, right), distances)
            distances = self._nj_forge_new_distance_matrix((left, right), distances, new_distances)
            ## ancestor_filename = "(%s-%s)" % (left_name, right_name)
            internal_node_count = len(internal_node_list) + 1
            ancestor_filename = "internal_%d" % internal_node_count
        
            clade_ancestor = Bio.Phylo.PhyloXML.Clade(branch_length=0)
            clade_ancestor.name = ancestor_filename
            
            try:
                clade_left = name_clade_map[left_name]
            except KeyError:
                clade_left = Bio.Phylo.PhyloXML.Clade()
                clade_left.name = left_name
                name_clade_map[left_name] = clade_left
            finally:
                clade_left.branch_length = new_distances[left]
            
            try:
                clade_right = name_clade_map[right_name]
            except KeyError:
                clade_right = Bio.Phylo.PhyloXML.Clade()
                clade_right.name = right_name
                name_clade_map[right_name] = clade_right        
            finally:
                clade_right.branch_length = new_distances[right]
            
            clade_ancestor.clades = [clade_left, clade_right]
            name_clade_map[ancestor_filename] = clade_ancestor
            internal_node_list.append(clade_ancestor)
            
            sequence_files.remove(left_name)
            sequence_files.remove(right_name)
            sequence_files.append(ancestor_filename)
            
            self.log.debug("Remaining number of sequences: %d" % len(sequence_files))
        
        ## here we have exactly 2 sequences left
        last_index = (sequence_files.index(ancestor_filename) + 1) % 2
        last_name = sequence_files[last_index]
        
        try:
            clade_last = name_clade_map[last_name]
        except KeyError:
            clade_last = Bio.Phylo.PhyloXML.Clade()
            clade_last.name = last_name
            name_clade_map[last_name] = clade_last        
        finally:
            clade_last.branch_length = new_distances[last_index]
        
        clade_ancestor.clades.append(clade_last) ## this one has 3 descendants
        
        self.tree = Bio.Phylo.PhyloXML.Phylogeny(root = clade_ancestor)
        self.internal_nodes = internal_node_list
    
    def _nj_create_q_matrix(self, D):
        """ computes the values of a neighbour-joining Q matrix out of the distance matrix D """
        Q = (D.shape[0]-2) * D - np.sum(D, axis=0, keepdims=True) - np.sum(D, axis=1, keepdims=True)
        
        return Q
    
    def _nj_nearest_neighbours(self, D):
        """ returns the nearest neighbours from the distance matrix D """
        np.fill_diagonal(D, np.inf)
        (min_i, min_j) = np.unravel_index(np.argmin(D), D.shape)
    
        return (min_i, min_j)
    
    def _nj_get_distances_to_new_node(self, left_right_tuple, D):
        (left, right) = left_right_tuple
        r = D.shape[0]
        distances = np.zeros((r))
        
        ## first the dist from the joined nodes to the new node
        d_left = 0.5 * D[left, right] + 1.0 / (2 * (r - 2)) * (np.sum(D[left, ]) - np.sum(D[right, ]))
        d_right = 0.5 * D[left, right] + 1.0 / (2 * (r - 2)) * (np.sum(D[right, ]) - np.sum(D[left, ]))

        ## now the dist from all other nodes to the new node
        distances = 0.5 * (D[left, :] - d_left) + 0.5 * (D[right, :] - d_right)

        distances[left] = d_left
        distances[right] = d_right

        return distances 
    
    def _nj_forge_new_distance_matrix(self, left_right_tuple, D, new_distances):
        (left, right) = left_right_tuple 
        r = D.shape[0]
        idx = np.repeat(True, r)
        idx[left] = False
        idx[right] = False
        
        E = D[idx][...,idx].copy()
    
        F = np.append(E, [new_distances[idx]], 0)
        G = np.append(F.transpose(), [np.append(new_distances[idx], [0])], 0).transpose()
        
        return G






import copy
import logging
import os
from itertools import combinations

import Bio
import fstlib
import numpy as np
import pandas as pd
import pyranges as pr
from joblib import Parallel, delayed

import medicc
from medicc import io, nj, tools

# prepare logger 
logger = logging.getLogger(__name__)


def main(input_df,
         asymm_fst,
         normal_name='diploid',
         input_tree=None,
         ancestral_reconstruction=True,
         chr_separator='X',
         prune_weight=0,
         allele_columns=['cn_a', 'cn_b'],
         wgd_x2=False,
         no_wgd=False,
         total_cn=False,
         n_cores=None):
    """ MEDICC Main Method """

    symbol_table = asymm_fst.input_symbols()

    ## Validate input
    logger.info("Validating input.")
    io.validate_input(input_df, symbol_table)

    ## Compile input data into FSAs stored in dictionaries
    logger.info("Compiling input sequences into FSAs.")
    FSA_dict = create_standard_fsa_dict_from_data(input_df, symbol_table, chr_separator)
    sample_labels = input_df.index.get_level_values('sample_id').unique()

    ## Reconstruct a tree
    if input_tree is None:
        ## Calculate pairwise distances
        logger.info("Calculating pairwise distance matrices")
        if n_cores is not None and n_cores > 1:
            pairwise_distances = parallelization_calc_pairwise_distance_matrix(sample_labels, 
                                                                        asymm_fst,
                                                                        FSA_dict,
                                                                        n_cores)
        else:
            pairwise_distances = calc_pairwise_distance_matrix(asymm_fst, FSA_dict)

        logger.info("Inferring tree topology.")
        nj_tree = infer_tree_topology(
            pairwise_distances.values, pairwise_distances.index, normal_name=normal_name)
    else:
        logger.info("Tree provided, using it. No pairwise distance matrix is calculated!")

        pairwise_distances = pd.DataFrame(0, columns=FSA_dict.keys(), index=FSA_dict.keys())

        assert len([x for x in list(input_tree.find_clades()) if x.name is not None and 'internal' not in x.name]) == \
            len(np.unique(input_df.index.get_level_values('sample_id'))), \
            "Number of samples differs in input tree and input dataframe"
        assert np.all(
            np.sort([x.name for x in list(input_tree.find_clades()) if x.name is not None and 'internal' not in x.name]) ==
            np.sort(np.unique(input_df.index.get_level_values('sample_id')))), \
            "Input tree does not match input dataframe"
        
        # necessary for the way that reconstruct_ancestors is performed
        if ancestral_reconstruction:
            input_tree.root_with_outgroup([x for x in input_tree.root.clades if x.name != normal_name][0].name)

        nj_tree = input_tree


    tools.set_sequences_on_tree_from_df(nj_tree, input_df)

    final_tree = copy.deepcopy(nj_tree)

    if ancestral_reconstruction:
        logger.info("Reconstructing ancestors.")
        ancestors = medicc.reconstruct_ancestors(tree=final_tree,
                                                 samples_dict=FSA_dict,
                                                 fst=asymm_fst,
                                                 normal_name=normal_name,
                                                 prune_weight=prune_weight)

        ## Create and write output data frame with ancestors
        logger.info("Creating output copynumbers.")
        output_df = create_df_from_fsa(input_df, ancestors)

        ## Update branch lengths with ancestors
        logger.info("Updating branch lengths of final tree using ancestors.")
        tools.set_sequences_on_tree_from_df(final_tree, output_df)
        update_branch_lengths(final_tree, asymm_fst, ancestors, normal_name)

    nj_tree.root_with_outgroup(normal_name)
    final_tree.root_with_outgroup(normal_name)

    if ancestral_reconstruction:
        output_df, events_df = calculate_all_cn_events(
            final_tree, output_df, allele_columns, normal_name,
            wgd_x2=wgd_x2, no_wgd=no_wgd, total_cn=total_cn)
        if len(events_df) != final_tree.total_branch_length():
            faulty_nodes = []
            for node in final_tree.find_clades():
                if node.name is not None and node.name != normal_name and node.branch_length != len(events_df.loc[node.name]):
                    faulty_nodes.append(node.name)
            logger.warn("Event recreation was faulty. Events in '_cn_events_df.tsv' will be "
                        f"incorrect for the following nodes: {faulty_nodes}")
    else:
        events_df = None
        output_df = input_df


    return sample_labels, pairwise_distances, nj_tree, final_tree, output_df, events_df


def create_standard_fsa_dict_from_data(input_data,
                                       symbol_table: fstlib.SymbolTable,
                                       separator: str = "X") -> dict:
    """ Creates a dictionary of FSAs from input DataFrame or Series.
    The keys of the dictionary are the sample/taxon names. 
    If the input is a DataFrame, the FSA will be the concatenated copy number profiles of all allele columns"""

    fsa_dict = {}
    if isinstance(input_data, pd.DataFrame):
        logger.info('Creating FSA for pd.DataFrame with the following data columns:\n{}'.format(
            input_data.columns))
        def aggregate_copy_number_profile(copy_number_profile):
            return separator.join([separator.join(["".join(x.astype('str'))
                                                   for _, x in cnp[allele].groupby('chrom')]) for allele in copy_number_profile.columns])

    elif isinstance(input_data, pd.Series):
        logger.info('Creating FSA for pd.Series with the name {}'.format(input_data.name))
        def aggregate_copy_number_profile(copy_number_profile):
            return separator.join(["".join(x.astype('str')) for _, x in copy_number_profile.groupby('chrom')])

    else:
        raise MEDICCError("Input to function create_standard_fsa_dict_from_data has to be either"
                          "pd.DataFrame or pd.Series. \n input provided was {}".format(type(input_data)))
    
    for taxon, cnp in input_data.groupby('sample_id'):
        cn_str = aggregate_copy_number_profile(cnp)
        fsa_dict[taxon] = fstlib.factory.from_string(cn_str,
                                                     arc_type="standard",
                                                     isymbols=symbol_table,
                                                     osymbols=symbol_table)

    return fsa_dict


def create_phasing_fsa_dict_from_df(input_df: pd.DataFrame, symbol_table: fstlib.SymbolTable, separator: str = "X") -> dict:
    """ Creates a dictionary of FSAs from two allele columns (Pandas DataFrame).
    The keys of the dictionary are the sample/taxon names. """
    allele_columns = input_df.columns
    if len(allele_columns) != 2:
        raise MEDICCError("Need exactly two alleles for phasing.")

    fsa_dict = {}
    for taxon, cnp in input_df.groupby('sample_id'):
        allele_a = cnp[allele_columns[0]]
        allele_b = cnp[allele_columns[1]]
        cn_str_a = separator.join(["".join(x) for _,x in allele_a.groupby(level='chrom', sort=False)])
        cn_str_b = separator.join(["".join(x) for _,x in allele_b.groupby(level='chrom', sort=False)])
        encoded = np.array([list(zip(cn_str_a, cn_str_b)), list(zip(cn_str_b, cn_str_a))])
        fsa_dict[taxon] = fstlib.factory.from_array(encoded, symbols=symbol_table, arc_type='standard')
        fsa_dict[taxon] = fstlib.determinize(fsa_dict[taxon]).minimize()

    return fsa_dict

def phase(input_df: pd.DataFrame, model_fst: fstlib.Fst, reference_sample='diploid', separator: str = 'X') -> pd.DataFrame:
    """ Phases every FST against the reference sample. 
    Returns two standard FSA dicts, one for each allele. """
    
    diploid_fsa = medicc.tools.create_diploid_fsa(model_fst)
    phasing_dict = medicc.create_phasing_fsa_dict_from_df(input_df, model_fst.input_symbols(), separator)
    fsa_dict_a, fsa_dict_b, _ = phase_dict(phasing_dict, model_fst, diploid_fsa)
    output_df = medicc.create_df_from_phasing_fsa(input_df, [fsa_dict_a, fsa_dict_b], separator)

    return output_df

def phase_dict(phasing_dict, model_fst, reference_fst):
    """ Phases every FST against the reference sample. 
    Returns two standard FSA dicts, one for each allele. """
    fsa_dict_a = {}    
    fsa_dict_b = {}
    scores = {}
    left = (reference_fst * model_fst).project('output')
    right = (~model_fst * reference_fst).project('input')
    for sample_id, sample_fst in phasing_dict.items():
        phased_fst = fstlib.align(sample_fst, left, right).topsort()
        score = fstlib.shortestdistance(phased_fst, reverse=True)[phased_fst.start()]
        scores[sample_id] = float(score)
        fsa_dict_a[sample_id] = fstlib.arcmap(phased_fst.copy().project('input'), map_type='rmweight')
        fsa_dict_b[sample_id] = fstlib.arcmap(phased_fst.project('output'), map_type='rmweight')
    
    return fsa_dict_a, fsa_dict_b, scores


def create_df_from_fsa(input_df: pd.DataFrame, fsa, separator: str = 'X'):
    """ 
    Takes a single FSA dict or a list of FSA dicts and extracts the copy number profiles.
    The allele names are taken from the input_df columns and the returned data frame has the same 
    number of rows and row index as the input_df. """

    alleles = input_df.columns
    if not isinstance(fsa, dict):
        raise MEDICCError("fsa input to create_df_from_fsa has to be a dict"
                          "Input type is {}".format(type(fsa)))

    nr_alleles = len(alleles)
    samples = input_df.index.get_level_values('sample_id').unique()
    output_df = input_df.unstack('sample_id')

    # Create dict and concat later to prevent pandas PerformanceWarning
    internal_cns = dict()
    for node in fsa:
        if node in samples:
            continue
        cns = tools.fsa_to_string(fsa[node]).split(separator)
        if len(cns) % nr_alleles != 0:
            raise MEDICCError('For sample {} we have {} haplotype-specific chromosomes for {} alleles'
                              '\nnumber of chromosomes has to be divisible by nr of alleles'.format(node,
                                                                                                    len(cns),
                                                                                                    nr_alleles))
        nr_chroms = int(len(cns) // nr_alleles)
        for i, allele in enumerate(alleles):
            cn = list(''.join(cns[(i*nr_chroms):((i+1)*nr_chroms)]))
            internal_cns[(allele, node)] = cn

    output_df = (pd.concat([output_df, pd.DataFrame(internal_cns, index=output_df.index)], axis=1)
                 .stack('sample_id')
                 .reorder_levels(['sample_id', 'chrom', 'start', 'end'])
                 .sort_index())

    return output_df


def create_df_from_phasing_fsa(input_df: pd.DataFrame, fsas, separator: str = 'X'):
    """ 
    Takes a two FSAs dicts from phasing and extracts the copy number profiles.
    The allele names are taken from the input_df columns and the returned data frame has the same 
    number of rows and row index as the input_df. """

    alleles = input_df.columns
    if len(fsas) != 2:
        raise MEDICCError("fsas has to be of length 2")
    if not all([isinstance(fsa, dict) for fsa in fsas]):
        raise MEDICCError("all fsas entries have to be dicts")
    if fsas[0].keys() != fsas[1].keys():
        raise MEDICCError("fsas keys have to be the same")


    output_df = input_df.copy()[[]]
    output_df[alleles] = ''

    for sample in fsas[0].keys():
        cns_a = tools.fsa_to_string(fsas[0][sample]).split(separator)
        cns_b = tools.fsa_to_string(fsas[1][sample]).split(separator)
        if len(cns_a) != len(cns_b):
            raise MEDICCError(f"length of alleles is not the same for sample {sample}")

        output_df.loc[sample, alleles[0]] = list(''.join(cns_a))
        output_df.loc[sample, alleles[1]] = list(''.join(cns_b))

    # output_df = output_df.stack('sample_id')
    # output_df = output_df.reorder_levels(['sample_id', 'chrom', 'start', 'end']).sort_index()
    
    return output_df


def parallelization_calc_pairwise_distance_matrix(sample_labels, asymm_fst, FSA_dict, n_cores):
    parallelization_groups = medicc.tools.create_parallelization_groups(len(sample_labels))
    parallelization_groups = [sample_labels[group] for group in parallelization_groups]
    logger.info("Running {} parallel runs on {} cores".format(len(parallelization_groups), n_cores))

    parallel_pairwise_distances = Parallel(n_jobs=n_cores)(delayed(calc_pairwise_distance_matrix)(
        asymm_fst, {key: val for key, val in FSA_dict.items() if key in cur_group}, True)
            for cur_group in parallelization_groups)

    pdm = medicc.tools.total_pdm_from_parallel_pdms(sample_labels, parallel_pairwise_distances)

    return pdm


def calc_pairwise_distance_matrix(model_fst, fsa_dict, parallel_run=True):
    '''Given a symmetric model FST and input FSAs in a form of a dictionary, output pairwise distance matrix'''

    samples = list(fsa_dict.keys())
    pdm = pd.DataFrame(0, index=samples, columns=samples, dtype=float)
    combs = list(combinations(samples, 2))
    ncombs = len(combs)

    for i, (sample_a, sample_b) in enumerate(combs):
        cur_dist = float(fstlib.kernel_score(model_fst, fsa_dict[sample_a], fsa_dict[sample_b]))
        pdm[sample_a][sample_b] = cur_dist
        pdm[sample_b][sample_a] = cur_dist

        if not parallel_run and (100*(i+1)/ncombs) % 10 == 0:  # log every 10%
            logger.info('%.2f%%', (i+1)/ncombs * 100)

    return pdm


def infer_tree_topology(pairwise_distances, labels, normal_name):
    if len(labels) > 2:
        tree = nj.NeighbourJoining(pairwise_distances, labels).tree

        tmpsearch = [c for c in tree.find_clades(name = normal_name)]
        normal_node = tmpsearch[0]
        root_path = tree.get_path(normal_node)[::-1]

        if len(root_path)>1:
            new_root = root_path[1]
            tree.root_with_outgroup(new_root)
    else:
        clade_ancestor = Bio.Phylo.PhyloXML.Clade(branch_length=0, name='internal_1')
        clade_ancestor.clades = [Bio.Phylo.PhyloXML.Clade(
            name=label, branch_length=0 if label == normal_name else 1) for label in labels]

        tree = Bio.Phylo.PhyloXML.Phylogeny(root=clade_ancestor)
        tree.root_with_outgroup(normal_name)

    return tree


def update_branch_lengths(tree, fst, ancestor_fsa, normal_name='diploid'):
    """ Updates the branch lengths in the tree using the internal nodes supplied in the FSA dict 
    """
    if len(ancestor_fsa) == 2:
        child_clade = [x for x in tree.find_clades() if x.name is not None and x.name != normal_name][0]
        child_clade.branch_length = float(fstlib.score(
            fst, ancestor_fsa[normal_name], ancestor_fsa[child_clade.name]))

    if not isinstance(ancestor_fsa, dict):
        raise MEDICCError("input ancestor_fsa to function update_branch_lengths has to be either a dict"
                          "provided type is {}".format(type(ancestor_fsa)))

    def _distance_to_child(fst, ancestor_fsa, sample_1, sample_2):
        return float(fstlib.score(fst, ancestor_fsa[sample_1], ancestor_fsa[sample_2]))

    for clade in tree.find_clades():
        if clade.name is None:
            continue
        children = clade.clades
        if len(children) != 0:
            for child in children:
                if child.name == normal_name:  # exception: evolution goes from diploid to internal node
                    brs = _distance_to_child(fst, ancestor_fsa, child.name, clade.name)
                else:
                    brs = _distance_to_child(fst, ancestor_fsa, clade.name, child.name)
                child.branch_length = brs


def calculate_all_cn_events(tree, cur_df, alleles=['cn_a', 'cn_b'], normal_name='diploid',
                            wgd_x2=False, no_wgd=False, total_cn=False):
    """Create a DataFrame containing all copy-number events in the current data

    Args:
        tree (Bio.Phylo.Tree): Phylogenetic tree created by MEDICC2's tree reconstruction
        cur_df (pandas.DataFrame): DataFrame containing the copy-numbers of the samples and internal nodes
        alleles (list, optional): List of alleles. Defaults to ['cn_a', 'cn_b'].
        normal_name (str, optional): Name of the normal sample. Defaults to 'diploid'.

    Returns:
        pandas.DataFrame: Updated copy-number DataFrame
        pandas.DataFrame: DataFrame of copy-number events
    """
    
    cur_df[['is_gain', 'is_loss', 'is_wgd']] = False
    cur_df[alleles] = cur_df[alleles].astype(int)
    if tree == None:
        cur_df[['is_normal', 'is_clonal']] = False
        events = None
    else:

        events = pd.DataFrame(columns=['sample_id', 'chrom', 'start',
                                    'end', 'allele', 'type', 'cn_child'])

        clades = [x for x in tree.find_clades()]

        for clade in clades:
            if not len(clade.clades):
                continue
            if clade.name is None:
                clade = copy.deepcopy(clade)
                clade.name = normal_name
            for child in clade.clades:
                if child.branch_length == 0:
                    continue

                cur_df, cur_events = calculate_cn_events_per_branch(
                    cur_df, clade.name, child.name, alleles=alleles, wgd_x2=wgd_x2,
                    total_cn=total_cn, no_wgd=no_wgd, normal_name=normal_name)

                events = pd.concat([events, cur_events])

        events = events.reset_index(drop=True)

        is_normal = ~cur_df.unstack('sample_id')[['is_loss', 'is_gain', 'is_wgd']].any(axis=1)
        is_normal.name = 'is_normal'
        mrca = [x for x in tree.root.clades if x.name != normal_name][0].name
        is_clonal = ~cur_df.loc[cur_df.index.get_level_values('sample_id')!=mrca].unstack('sample_id')[['is_loss', 'is_gain', 'is_wgd']].any(axis=1)
        is_clonal.name = 'is_clonal'

        cur_df = cur_df.drop(['is_normal', 'is_clonal'], axis=1, errors='ignore')
        cur_df = (cur_df
                .join(is_normal, how='inner')
                .reorder_levels(['sample_id', 'chrom', 'start', 'end'])
                .sort_index()
                .join(is_clonal, how='inner')
                .reset_index())
        cur_df['chrom'] = tools.format_chromosomes(cur_df['chrom'])
        cur_df = (cur_df
                .set_index(['sample_id', 'chrom', 'start', 'end'])
                .sort_index())

        events = events.set_index(['sample_id', 'chrom', 'start', 'end'])

    return cur_df, events


def calculate_cn_events_per_branch(cur_df, parent_name, child_name, alleles=['cn_a', 'cn_b'],
                                   wgd_x2=False, total_cn=False, no_wgd=False, normal_name='diploid'):
    """Calculate copy-number events for a single branch. Used in calculate_all_cn_events

    Args:
        cur_df (pandas.DataFrame): DataFrame containing the copy-numbers of the samples and internal nodes
        parent_name (str): Name of the parent sample
        child_name (str): Name of the child sample
        alleles (list, optional): List of alleles. Defaults to ['cn_a', 'cn_b'].

    Returns:
        pandas.DataFrame: Updated copy-number DataFrame
        pandas.DataFrame: DataFrame of copy-number events
    """

    cur_df = cur_df.copy()
    if len(np.setdiff1d(['is_gain', 'is_loss', 'is_wgd'], cur_df.columns)) > 0:
        cur_df[['is_gain', 'is_loss', 'is_wgd']] = False
    cur_df[alleles] = cur_df[alleles].astype(int)

    # TODO: load these outside of the function so they are not loaded every time
    asymm_fst, asymm_fst_nowgd, asymm_fst_1_wgd, asymm_fst_2_wgd, symbol_table = io.load_main_fsts(
        return_symbol_table=True)
    if wgd_x2:
        asymm_fst = io.read_fst(wgd_x2=True)
        asymm_fst_1_wgd = io.read_fst(wgd_x2=True, n_wgd=1)
        asymm_fst_2_wgd = None
    if no_wgd:
        asymm_fst = io.read_fst(no_wgd=True)
        asymm_fst_1_wgd = None
        asymm_fst_2_wgd = None


    events_df = pd.DataFrame(columns=['sample_id', 'chrom', 'start', 'end', 'allele', 'type', 'cn_child'])

    cur_parent_cn = cur_df.loc[parent_name, alleles].astype(int)
    cur_child_cn = cur_df.loc[child_name, alleles].astype(int)

    def get_int_chrom(x):
        if x == 'chrX':
            return 23
        elif x == 'chrY':
            return 24
        else:
            return int(x.split('chr')[-1])

    cur_chroms = cur_df.loc[normal_name].index.get_level_values(
        'chrom').map(get_int_chrom).values.astype(int)

    # 1. find total loss (loh)
    parent_loh = cur_parent_cn == 0
    for allele in alleles:

        cur_loh = cur_child_cn.loc[~parent_loh[allele], allele] == 0
        if cur_loh.sum() == 0:
            continue

        cur_df.loc[child_name, 'is_loss'] = (cur_df.loc[child_name, 'is_loss'].values 
                                             + np.logical_and(cur_child_cn[allele] == 0, 
                                                              cur_parent_cn[allele] != 0).values)

        max_previous_cn = np.max(
            np.unique(cur_parent_cn.loc[~parent_loh[allele], allele].loc[cur_loh]))

        for _ in np.arange(max_previous_cn):
            cur_loh_and_parental_val = np.logical_and(cur_loh.values, 
                                                        cur_parent_cn.loc[~parent_loh[allele], allele] > 0).values

            # + cur_chroms enables detection of chromosome boundaries
            event_labels_ = ((np.cumsum(np.concatenate([[0], np.diff(
                (cur_loh_and_parental_val + cur_chroms[~parent_loh[allele]]))])
                * cur_loh_and_parental_val) + 1)
                * cur_loh_and_parental_val)

            # Label events starting at 1
            event_labels = np.zeros_like(event_labels_)
            for i, j in enumerate(np.setdiff1d(np.unique(event_labels_), [0])):
                event_labels[event_labels_ == j] = i + 1

            cur_parent_cn.loc[parent_loh.loc[~parent_loh[allele],
                                               allele].index[cur_loh_and_parental_val], allele] -= 1

            cur_events = (cur_parent_cn
                        .loc[~parent_loh[allele]]
                        .reset_index()
                        .loc[np.array([np.argmax(event_labels == ind) for ind in np.setdiff1d(np.unique(event_labels), [0])])]
                        [['chrom', 'start', 'end']].values)
            # adjust ends
            cur_events[:, 2] = (cur_parent_cn
                                .loc[~parent_loh[allele]]
                                .reset_index()
                                .loc[np.array([len(event_labels) - np.argmax(event_labels[::-1] == ind) - 1 for ind in np.setdiff1d(np.unique(event_labels), [0])])]
                                ['end'].values)

            cur_ind = np.arange(len(events_df), len(events_df)+len(cur_events))
            events_df = pd.concat([events_df, pd.DataFrame(index=cur_ind)])
            events_df.loc[cur_ind, 'sample_id'] = child_name
            events_df.loc[cur_ind, 'allele'] = allele
            events_df.loc[cur_ind, 'type'] = 'loh'
            events_df.loc[cur_ind, 'cn_child'] = 0
            events_df.loc[cur_ind, ['chrom', 'start', 'end']] = cur_events[:, :3]

            # recalculate parental_loss and cur_loh for next iteration
            parent_loh = cur_parent_cn <= 0
            cur_loh = cur_child_cn.loc[~parent_loh[allele], allele] == 0

        cur_parent_cn.loc[cur_parent_cn[allele] < 0, allele] = 0
    loh_pos = (cur_parent_cn == 0)

    # 2. WGDs
    # only check if >30% of is gained
    wgd_candidate_threshold = 0.3

    widths = cur_df.loc[[child_name]].eval('end-start')
    fraction_gain = ((cur_df.loc[child_name, alleles] > 1).astype(int).sum(axis=1) * widths.loc[child_name]
                        ).sum() / (2 * widths.loc[child_name].sum())
    parent_fsa = fstlib.factory.from_string('X'.join(['X'.join(["".join(x.astype('str')) for _, x in cur_df.loc[parent_name, alleles][allele].groupby('chrom')]) for allele in alleles]),
                                            arc_type="standard",
                                            isymbols=symbol_table,
                                            osymbols=symbol_table)
    child_fsa = fstlib.factory.from_string('X'.join(['X'.join(["".join(x.astype('str')) for _, x in cur_df.loc[child_name, alleles][allele].groupby('chrom')]) for allele in alleles]),
                                            arc_type="standard",
                                            isymbols=symbol_table,
                                            osymbols=symbol_table)

    score_wgd = float(fstlib.score(asymm_fst, parent_fsa, child_fsa))
    fraction_double_gain = (((cur_df.loc[child_name, alleles] > 2)
                            .astype(int)
                            .sum(axis=1)
                            * widths.loc[child_name]
                            ).sum() / widths.loc[child_name].sum())
    if not no_wgd and fraction_gain > wgd_candidate_threshold:
        if wgd_x2:
            # double wgd
            if (fraction_double_gain > wgd_candidate_threshold) and (float(fstlib.score(asymm_fst_1_wgd, parent_fsa, child_fsa)) != score_wgd):
                cur_parent_cn = 4 * cur_parent_cn
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True
            # single wgd
            elif float(fstlib.score(asymm_fst_nowgd, parent_fsa, child_fsa)) != score_wgd:
                cur_parent_cn = 2 * cur_parent_cn
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True

        elif total_cn:
            # single wgd
            if float(fstlib.score(asymm_fst_nowgd, parent_fsa, child_fsa)) != score_wgd:
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 2
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True

        else:
            # triple wgd
            if (fraction_double_gain > wgd_candidate_threshold) and (float(fstlib.score(asymm_fst_2_wgd, parent_fsa, child_fsa)) != score_wgd):
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 3
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True
            # double wgd
            elif (fraction_double_gain > wgd_candidate_threshold) and (float(fstlib.score(asymm_fst_1_wgd, parent_fsa, child_fsa)) != score_wgd):
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 2
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                    cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True
            # single wgd
            elif float(fstlib.score(asymm_fst_nowgd, parent_fsa, child_fsa)) != score_wgd:
                cur_parent_cn[~loh_pos] = cur_parent_cn[~loh_pos] + 1
                events_df.loc[len(events_df.index)] = [child_name, 'chr0', cur_df.index.get_level_values('start').min(),
                                                cur_df.index.get_level_values('end').max(), 'both', 'wgd', 0]
                cur_df.loc[child_name, 'is_wgd'] = True

    # 3. losses and gains
    for allele in alleles:

        cn_changes = (cur_child_cn[allele] - cur_parent_cn[allele]).values
        all_cn_change_vals = np.unique(cn_changes)

        cur_df.loc[child_name, 'is_loss'] = np.logical_or(cur_df.loc[child_name, 'is_loss'].values,
                                                          cn_changes < 0)
        cur_df.loc[child_name, 'is_gain'] = np.logical_or(cur_df.loc[child_name, 'is_gain'].values,
                                                          cn_changes > 0)

        # enumerate over all possible change values
        all_cn_change_vals = np.setdiff1d(np.arange(np.min([np.min(all_cn_change_vals), 0]), np.max(all_cn_change_vals)+1), [0])
        for cur_cn_change in all_cn_change_vals[np.argsort(np.abs(all_cn_change_vals))[::-1]]:
            cur_event = 'gain' if cur_cn_change > 0 else 'loss'

            cur_change_location = ((cur_child_cn.loc[~loh_pos[allele], allele] -
                        cur_parent_cn.loc[~loh_pos[allele], allele]) == cur_cn_change)

            # + cur_chroms enables detection of chromosome boundaries
            event_labels_ = ((np.cumsum(np.concatenate([[0], np.diff(
                (cur_change_location.values + cur_chroms[~loh_pos[allele]]))])
                * cur_change_location.values) + 1)
                * cur_change_location.values)

            # Label events starting at 1
            event_labels = np.zeros_like(event_labels_)
            for i, j in enumerate(np.setdiff1d(np.unique(event_labels_), [0])):
                event_labels[event_labels_ == j] = i + 1

            cur_events = (cur_child_cn
                          .loc[~loh_pos[allele]]
                          .reset_index()
                          .loc[np.array([np.argmax(event_labels == val) for val in np.setdiff1d(np.unique(event_labels), [0])])]
                          [['chrom', 'start', 'end', allele]].values)

            # adjust ends
            cur_events[:, 2] = (cur_child_cn
                                .loc[~loh_pos[allele]]
                                .reset_index()
                                .loc[np.array([len(event_labels) - np.argmax(event_labels[::-1] == val) - 1 for val in np.setdiff1d(np.unique(event_labels), [0])])]
                                ['end'].values)

            cur_ind = np.arange(len(events_df), len(events_df)+len(cur_events))
            events_df = pd.concat([events_df, pd.DataFrame(index=cur_ind)])
            events_df.loc[cur_ind, 'sample_id'] = child_name
            events_df.loc[cur_ind, 'allele'] = allele
            events_df.loc[cur_ind, 'type'] = cur_event
            events_df.loc[cur_ind, 'cn_child'] = cur_events[:, 3]
            events_df.loc[cur_ind, ['chrom', 'start', 'end']] = cur_events[:, :3]

            cur_child_cn.loc[np.intersect1d(loh_pos.loc[~loh_pos[allele]].index,
                                            cur_change_location.loc[cur_change_location].index), allele] += (1 if (cur_cn_change < 0) else -1)

    events_df['chrom'] = tools.format_chromosomes(events_df['chrom'])
    events_df = (events_df[['sample_id', 'allele', 'chrom', 'start', 'end', 'type', 'cn_child']]
                 .reset_index(drop=True)
                 .sort_values(['sample_id', 'allele', 'chrom', 'start', 'end', 'type', 'cn_child']))

    return cur_df, events_df


def compute_cn_change(df, tree, normal_name='diploid'):
    """Compute the copy-number changes per segment in all branches

    Args:
        df (pandas.DataFrame): DataFrame containing the copy-numbers of samples and internal nodes
        tree (Bio.Phylo.Tree): Phylogenetic tree
        normal_name (str, optional): Name of normal sample. Defaults to 'diploid'.

    Returns:
        pandas.DataFrame: DataFrame containing the copy-number changes
    """    
    cn_change = df.copy()
    alleles = cn_change.columns
    for allele in alleles:
        cn_change[allele] = cn_change[allele].astype('int')

    clades = [clade for clade in tree.find_clades(order = "postorder") if clade.name is not None and clade.name != normal_name]
    for clade in clades:
        for child in clade.clades:
            cn_change.loc[child.name, alleles] = cn_change.loc[child.name, alleles].values - cn_change.loc[clade.name, alleles].values
    cn_change.loc[clades[-1].name, alleles] = cn_change.loc[clades[-1].name, alleles].values - cn_change.loc[normal_name, alleles].values
    cn_change.loc[normal_name, alleles] = 0

    return cn_change


def summarize_patient(tree, pdm, sample_labels, normal_name='diploid', events_df=None):
    """Calculate several summary values for the provided samples

    Args:
        tree (Bio.Phylo.Tree): Phylogenetic tree
        pdm (pandas.DataFrame): Pairwise distance matrix between the samples
        sample_labels (list): List of all samples
        normal_name (str, optional): Name of normal sample. Defaults to 'diploid'.
        events_df (pandas.DataFrame, optional): DataFrame containg all copy-number events. Defaults to None.

    Returns:
        pandas.DataFrame: Summary DataFrame
    """    
    branch_lengths = []
    for parent in tree.find_clades(terminal=False, order="level"):
        for child in parent.clades:
            if child.branch_length:
                branch_lengths.append(child.branch_length)
    
    nsamples = len(sample_labels)
    tree_length = np.sum(branch_lengths)
    avg_branch_length = np.mean(branch_lengths)
    min_branch_length = np.min(branch_lengths)
    max_branch_length = np.max(branch_lengths)
    median_branch_length = np.median(branch_lengths)
    # p_star = stats.star_topology_test(pdm)
    # p_clock = stats.molecular_clock_test(pdm,
    #                                      np.flatnonzero(np.array(sample_labels) == normal_name)[0])
    if events_df is None:
        wgd_status = "unknown"
    else:
        if "wgd" in events_df['type'].values:
            wgd_status = "WGD on branch " + \
                "and ".join(events_df.loc[events_df['type'] ==
                                          'wgd'].index.get_level_values('sample_id'))
        else:
            wgd_status = "no WGD"

    result = pd.Series({
        'nsamples': nsamples,
        'normal_name': normal_name,
        'tree_length': tree_length,
        'mean_branch_length': avg_branch_length,
        'median_branch_length': median_branch_length,
        'min_branch_length': min_branch_length,
        'max_branch_length': max_branch_length,
        # 'p_star': p_star,
        # 'p_clock': p_clock,
        'wgd_status': wgd_status,
    })
    
    return result


def overlap_events(events_df=None, output_df=None, tree=None, overlap_threshold=0.9,
                   chromosome_bed='default', regions_bed='default',
                   replace_loh_with_loss=True, alleles=['cn_a', 'cn_b'],
                   replace_both_arms_with_chrom=True, normal_name='diploid'):
    """Overlap copy-number events with regions of interest

    Args:
        events_df (pandas.DataFrame, optional): All copy-number events. Defaults to None.
        output_df (pandas.DataFrame, optional): DataFrame containing all copy-numbers. Defaults to None.
        tree (Bio.Phylo.Tree, optional): Phylogenetic tree. Defaults to None.
        overlap_threshold (float, optional): Threshold above which an overlap is considered. Defaults to 0.9.
        chromosome_bed (str, optional): Name of BED file containing chromosome arm information. Defaults to 'default'.
        regions_bed (str, optional): Name of BED file containing regions of interest. Defaults to 'default'.
        replace_loh_with_loss (bool, optional): If True, loh is considered like a normal loss. Defaults to True.
        alleles (list, optional): List of alleles. Defaults to ['cn_a', 'cn_b'].
        replace_both_arms_with_chrom (bool, optional): If True, an event in the p- and q-arm of a chromosome will be displayed as a single event. Defaults to True.
        normal_name (str, optional): Name of normal sample. Defaults to 'diploid'.

    Returns:
        pandas.DataFrame: DataFrame with events concerning the regions of interest
    """                   

    if chromosome_bed == 'default':
        chromosome_bed = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                      "objects", "hg38_chromosome_arms.bed")
    if regions_bed == 'default':
        regions_bed = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   "objects", "Davoli_2013_TSG_OG_genes.bed")

    all_events = pd.DataFrame(columns=['Chromosome', 'Start', 'End', 'name', 'NumberOverlaps',
                                       'FractionOverlaps', 'event', 'branch']).set_index(['Chromosome', 'Start', 'End'])

    if events_df is None:
        if output_df is None or tree is None:
            raise MEDICCError("Either events_df or df and tree has to be specified")
        _, events_df = calculate_all_cn_events(tree, output_df, alleles=alleles, normal_name=normal_name)
    if replace_loh_with_loss:
        events_df.loc[events_df['type'] == 'loh', 'type'] = 'loss'

    # Read chromosome regions and other regions
    if chromosome_bed is None and regions_bed is None:
        raise MEDICCError("Either chromosome_bed or regions_bed has to be specified")

    chr_arm_regions = None
    if chromosome_bed is not None:
        logger.debug(f'Overlap with chromosomes bed file {chromosome_bed}')
        chr_arm_regions = medicc.io.read_bed_file(chromosome_bed)
        whole_chromosome = chr_arm_regions.groupby('Chromosome').min()
        whole_chromosome['End'] = chr_arm_regions.groupby('Chromosome')['End'].max()
        whole_chromosome['name'] = whole_chromosome.index
        chr_arm_regions = pd.concat([chr_arm_regions, whole_chromosome.reset_index()]).sort_values('Chromosome')
        chr_arm_regions = pr.PyRanges(chr_arm_regions)

    regions = None
    if regions_bed is not None:
        logger.debug(f'Overlap with regions bed file {regions_bed}')
        regions = []
        if isinstance(regions_bed, list) or isinstance(regions_bed, tuple):
            for f in regions_bed:
                regions.append(pr.PyRanges(medicc.io.read_bed_file(f)))
        else:
            regions.append(pr.PyRanges(medicc.io.read_bed_file(regions_bed)))

    # Add WGD
    for ind, _ in events_df.loc[events_df['type'] == 'wgd'].iterrows():
        all_events.loc[('all', '0', '0')] = ['WGD', 1, 1., 'WGD', ind[0]]

    for cur_branch in events_df.index.get_level_values('sample_id').unique():
        for allele in alleles:
            cur_events_df = events_df.loc[cur_branch]
            cur_events_df = cur_events_df.loc[cur_events_df['allele']==allele]
            for event_type in ['gain', 'loss'] if replace_loh_with_loss else ['gain', 'loh', 'loss']:
                cur_events_ranges = pr.PyRanges(cur_events_df.loc[cur_events_df['type'] == event_type].reset_index(
                ).rename({'chrom': 'Chromosome', 'start': 'Start', 'end': 'End'}, axis=1))

                # Calculate chromosomal events
                if chr_arm_regions is not None:
                    chr_events = overlap_regions(
                        chr_arm_regions, cur_events_ranges, event_type, cur_branch, overlap_threshold)
                    # remove arms if the whole chromosome is in there
                    if replace_both_arms_with_chrom and len(chr_events) > 0:
                        chr_events = chr_events[~chr_events['name'].isin(np.concatenate(
                            [[name + 'p', name + 'q'] if ('q' not in name and 'p' not in name) else [] for name in chr_events['name']]))]
                    all_events = pd.concat([all_events, chr_events])

                # Calculate other events
                if regions is not None:
                    for region in regions:
                        chr_events = overlap_regions(
                            region, cur_events_ranges, event_type, cur_branch, overlap_threshold)
                        all_events = pd.concat([all_events, chr_events])

    all_events['final_name'] = all_events['name'].apply(lambda x: x.split(
        'chr')[-1]) + all_events['event'].apply(lambda x: ' +' if x == 'gain' else (' -' if x == 'loss' else (' 0' if x == 'loh' else '')))

    all_events = all_events.set_index(['branch', 'name']).drop('NumberOverlaps', axis=1)
    all_events = all_events.reset_index().set_index('branch')
    
    return all_events


def overlap_regions(region, cur_events_ranges, event, branch, overlap_threshold):

    cur_events_overlaps = region.coverage(cur_events_ranges).as_df()
    cur_events_overlaps = cur_events_overlaps.loc[cur_events_overlaps['FractionOverlaps']
                                                > overlap_threshold]
    cur_events_overlaps = cur_events_overlaps.set_index(['Chromosome', 'Start', 'End'])
    cur_events_overlaps['event'] = event
    cur_events_overlaps['branch'] = branch

    return cur_events_overlaps


def detect_wgd(input_df, sample, total_cn=False, wgd_x2=False):
    wgd_fst = io.read_fst(total_copy_numbers=total_cn, wgd_x2=wgd_x2)
    no_wgd_fst = io.read_fst(no_wgd=True)

    diploid_fsa = medicc.tools.create_diploid_fsa(no_wgd_fst)
    symbol_table = no_wgd_fst.input_symbols()
    fsa_dict = medicc.create_standard_fsa_dict_from_data(input_df.loc[[sample]],
                                                         symbol_table, 'X')

    distance_wgd = float(fstlib.score(wgd_fst, diploid_fsa, fsa_dict[sample]))
    distance_no_wgd = float(fstlib.score(no_wgd_fst, diploid_fsa, fsa_dict[sample]))

    return distance_wgd != distance_no_wgd



class MEDICCError(Exception):
    pass



import logging.config
import os

import yaml

import medicc.bootstrap
import medicc.io
import medicc.nj
import medicc.plot
import medicc.sim
import medicc.stats
import medicc.tools
from medicc.ancestors import reconstruct_ancestors
from medicc.core import *
from medicc.factory import *

with open(os.path.join(os.path.dirname(__file__), 'logging_conf.yaml'), 'rt') as f:
    config = yaml.safe_load(f.read())
logging.config.dictConfig(config)



""" Python class for finite-state transducers over the real or log semiring"""

import logging
import math
import os
import sys
import tempfile
from io import StringIO

import numpy as np
import pandas as pd

import fstlib
import fstlib.algos
from fstlib.cext import pywrapfst

logger = logging.getLogger(__name__)

class Fst:
    """ Class that describes a finite-state transducer.
    """
    
    def __init__(self, arc_type = fstlib.Semiring.TROPICAL):
        if isinstance(arc_type, pywrapfst.MutableFst):
            self.fst = arc_type
        elif isinstance(arc_type, Fst):
            self.fst = arc_type.fst
        else:
            self.fst = pywrapfst.VectorFst(arc_type=arc_type)

    def __repr__(self):
        return self.fst.__repr__()

    def _repr_png_(self):
        if self.num_states() > 50: ## don't draw, too large, too slow
            logger.info('Too many states to draw')
            return
        g = self.to_graphviz(width=12, height=14)
        g.format='png'
        return g.pipe()

    def _repr_svg_(self):
        if self.num_states() > 50: ## don't draw, too large, too slow
            logger.info('Too many states to draw')
            return
        g = self.to_graphviz(width=12, height=14)
        g.format='svg'
        return str(g.pipe(), 'utf-8')

    def __str__(self):
        return self.fst.__str__()

    def __mul__(self, other):
        return fstlib.compose(self.arcsort('olabel'), other)

    def __add__(self, other):
        return self.copy().union(other)

    def __invert__(self):
        return self.copy().invert()

    def add_arc(self, state, arc):
        """ Arc can be pywrapfst.Arc or 4-tuple (ilabel, olabel, weight, nextstate) """
        if isinstance(arc, pywrapfst.Arc):
            pass
        else: ## assume tuple / unpackable
            ### adds an arc object expecting the right weight type and trying
            ### to resolve symbolic (non-integer) labels via the symbol table.
            ilabel, olabel, weight, nextstate = arc
            isyms = self.input_symbols()
            osyms = self.output_symbols()
            if isinstance(ilabel, str):
                if isyms is not None:
                    isym = isyms.find(ilabel)
                else:
                    raise FSTError('ilabel is String but no input symbol table available.')
            else:
                isym = ilabel

            if isinstance(olabel, str):
                if osyms is not None:
                    osym = osyms.find(olabel)
                else:
                    raise FSTError('olabel is String but no output symbol table available.')
            else:
                osym = olabel

            if isinstance(weight, pywrapfst.Weight):
                pass
            else:
                weight = pywrapfst.Weight(self.weight_type(), weight)
            arc = pywrapfst.Arc(isym, osym, weight, nextstate)
        
        self.fst.add_arc(state, arc)
        return self

    def add_arcs(self, state, iterable_of_arcs):
        for arc in iterable_of_arcs:
            self.add_arc(state, arc)
        return self

    def add_state(self):
        return self.fst.add_state()

    def add_states(self, n):
        self.reserve_states(n)
        ids = np.array([self.add_state() for i in range(n)], dtype=np.int32)
        return ids
        
    def arc_type(self):
        return self.fst.arc_type()

    def arcs(self, state):
        return self.fst.arcs(state)

    def arcsort(self, sort_type='ilabel'):
        self.fst.arcsort(sort_type)
        return self

    def closure(self, closure_plus = False):
        self.fst.closure(closure_plus)
        return self

    def concat(self, ifst):
        self.fst.concat(ifst.fst)
        return self

    def connect(self):
        self.fst.connect()
        return self

    def copy(self):
        newfst = self.fst.copy()
        return fstlib.core.Fst(newfst)

    def decode(self, encoder):
        self.fst.decode(encoder)
        return self

    def delete_arcs(self, state, n=0):
        self.fst.delete_arcs(state, n)
        return self

    def delete_states(self, states=None):
        self.fst.delete_states(states)
        return self

    def draw(self, filename, isymbols=None, osymbols=None, ssymbols=None, acceptor=False, title="", 
          width=8.5, height=11, portrait=False, vertical=False, ranksep=0.4, nodesep=0.25, 
          fontsize=14, precision=5, float_format='g', show_weight_one=False):
        
        self.fst.draw(filename, isymbols, osymbols, ssymbols, acceptor, title, width, height, portrait, 
                vertical, ranksep, nodesep, fontsize, precision, float_format, show_weight_one)

    def encode(self, encoder):
        self.fst.encode(encoder)
        return self

    def final(self, state):
        return self.fst.final(state)

    def fst_type(self):
        return self.fst.fst_type()

    def info(self, name=None):
        df = fstlib.ext.info(self, name=name)
        return df

    def input_symbols(self):
        return self.fst.input_symbols()

    def invert(self):
        self.fst.invert()
        return self

    def is_final(self, state):
        return self.final(state) != pywrapfst.Weight.zero(self.weight_type())

    def minimize(self, delta=1e-6, allow_nondet=False):
        self.fst.minimize(delta, allow_nondet)
        return self

    def mutable_arcs(self, state):
        return self.fst.mutable_arcs(state)

    def mutable_input_symbols(self):
        return self.fst.mutable_input_symbols()

    def mutable_output_symbols(self):
        return self.fst.mutable_output_symbols()

    def num_arcs(self, state):
        return self.fst.num_arcs(state)

    def num_input_epsilons(self, state):
        return self.fst.num_input_epsilons(state)

    def num_output_epsilons(self, state):
        return self.fst.num_output_epsilons(state)

    def num_states(self):
        return self.fst.num_states()

    def output_symbols(self):
        return self.fst.output_symbols()

    def paths(self):
        pdfs = fstlib.algos.PathDepthFirstSearch(self)
        return pdfs.get_paths()

    def project(self, project_type='input'):
        self.fst.project(project_type)
        return self

    def properties(self, mask, test):
        return self.fst.properties(mask,test)

    def prune(self, delta=fstlib.DEF_DELTA, nstate=fstlib.NO_STATE_ID, weight=None):
        self.fst.prune(delta, nstate, weight)
        return self

    def push(self, delta=fstlib.DEF_DELTA, remove_total_weight=False, to_final=False):
        self.fst.push(delta, remove_total_weight, to_final)
        return self

    def relabel_pairs(self, ipairs=None, opairs=None):
        self.fst.relabel_pairs(ipairs, opairs)
        return self

    def relabel_tables(self, old_isymbols=None, new_isymbols=None, unknown_isymbols="", attach_new_isymbols=True,
                    old_osymbols=None, new_osymbols=None, unknown_osymbols="", attach_new_osymbols=True):
        self.fst.relabel_tables(old_isymbols, new_isymbols, unknown_isymbols, attach_new_isymbols,
                          old_osymbols, new_osymbols, unknown_osymbols, attach_new_osymbols)
        return self

    def reserve_arcs(self, state, n):
        return self.fst.reserve_arcs(state, n)

    def reserve_states(self, n):
        return self.fst.reserve_states(n)

    def reweight(self, potentials, to_final=False):
        self.fst.reweight(potentials, to_final)
        return self

    def rmepsilon(self, queue_type='auto', connect=True, weight=None, nstate=fstlib.NO_STATE_ID, delta=1e-6):
        self.fst.rmepsilon(queue_type, connect, weight, nstate, delta)
        return self

    def set_final(self, state, weight):
        self.fst.set_final(state, weight)
        return self

    def set_input_symbols(self, syms):
        self.fst.set_input_symbols(syms)
        return self

    def set_output_symbols(self, syms):
        self.fst.set_output_symbols(syms)
        return self

    def set_properties(self, props, mask):
        return self.fst.set_properties(props, mask)

    def set_start(self, state):
        self.fst.set_start(state)
        return self

    def start(self):
        return self.fst.start()

    def states(self):
        return self.fst.states()

    def text(self):
        return self.fst.text()

    def to_dataframe(self, select='arcs', to_real=False):
        if self.input_symbols() is not None:
            isyms = dict([(i, s) for i,s in self.input_symbols()])
        else:
            isyms = None
        if self.output_symbols() is not None:
            osyms = dict([(i, s) for i,s in self.output_symbols()])
        else:
            osyms = None

        def get_arc_record(state, arc):
            if isyms is not None:
                isym = isyms[arc.ilabel]
            else:
                isym = arc.ilabel
            if osyms is not None:
                osym = osyms[arc.olabel]
            else:
                osym = arc.olabel
            if to_real:
                weight = np.exp(-float(arc.weight))
            else:
                weight = float(arc.weight)
            record = (str(id(arc)), state, arc.nextstate, isym, osym, weight)
            return record

        if select == 'arcs':
            tuples = [get_arc_record(state, arc) for state in self.states() for arc in self.arcs(state)]
            df = pd.DataFrame.from_records(tuples, columns=['arcid', 'state_from', 'state_to', 'ilabel', 'olabel', 'weight'])
            df.set_index('arcid', inplace=True)
        elif select == 'states':
            tuples = [(state, np.exp(-float(self.final(state))) if to_real else float(self.final(state)), self.is_final(state)) for state in self.states()]
            index = [str(id(state)) for state in self.states()]
            df = pd.DataFrame.from_records(tuples, columns=['state', 'weight', 'is_final'], index=index)
        else:
            raise FSTError("Unknown select parameter, can be 'arcs' or 'states'")
        return df

    def to_graphviz(self, isymbols=None, osymbols=None, ssymbols=None, acceptor=False, title="", 
          width=8.5, height=11, portrait=True, vertical=False, ranksep=0.4, nodesep=0.25, 
          fontsize=14, precision=5, float_format='g', show_weight_one=False):
        import graphviz as gv
        
        tmpfile = tempfile.mktemp()
        self.draw(tmpfile, isymbols, osymbols, ssymbols, acceptor, title, width, height, portrait, 
                vertical, ranksep, nodesep, fontsize, precision, float_format, show_weight_one)
        with open(tmpfile, 'r') as fd:
            dot = fd.read()
        os.unlink(tmpfile)
        src = gv.Source(dot)
        return src

    def to_real(self):
        """ Shortcut for a constructive weight map from log to real, mainly
        for plotting and printing purposes. """
        return fstlib.weight_map(self, fstlib.algos.map_log_to_real)

    def topsort(self):
        self.fst.topsort()
        return self

    def union(self, ifst):
        self.fst.union(ifst.fst)
        return self

    def verify(self):
        return self.fst.verify()

    def weight_map(self, func, with_final_states=True):
        ## maps each arc and final state weight to a new value using the function provided """
        for state in self.states():
            final_weight = self.final(state)
            if final_weight != pywrapfst.Weight.zero(self.weight_type()) and with_final_states: ## final state
                self.set_final(state, func(final_weight))
            mai = self.mutable_arcs(state)
            for arc in mai:
                arc.weight = pywrapfst.Weight(self.weight_type(), func(arc.weight))
                mai.set_value(arc)

        return self

    def weight_type(self):
        return self.fst.weight_type()

    def write(self, filename):
        return self.fst.write(filename)

    def write_to_string(self):
        return self.fst.write_to_string()

    @classmethod
    def read(cls, source):
        ofst = pywrapfst.Fst.read(source)
        return Fst(ofst)

    @classmethod
    def read_from_string(cls, state):
        ofst = pywrapfst.Fst.read(state)
        return Fst(ofst)

class Path(list):
    """ simple extension of list class that includes a final weight """
    __attributes = ["finalWeight"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.finalWeight = kwargs.get("finalWeight", 0)
        valid = [p in self.__attributes for p in kwargs.keys()]
        if not all(valid):
            raise FSTError("Unknown attribute in [" + ",".join(kwargs.keys()) + "]. Allowed arguments are: [" + ",".join(self.__attributes) + "].")

    def copy(self):
        newpath = fstlib.Path(self)
        newpath.finalWeight = self.finalWeight
        return newpath

class FSTError(Exception):
    def __init__(self, message="<unknown error>"):
        self.message = message

    def __str__(self):
        return str(self.message)

## module scope variables
_last_seed = None

## function definitions
def _generate_seed():
    global _last_seed
    random_data = os.urandom(4) 
    seed = int.from_bytes(random_data, byteorder="big") 
    _last_seed = seed
    return seed

def arcmap(ifst, delta=fstlib.DEF_DELTA, map_type='identity', power=1.0, weight=None):
    newfst = pywrapfst.arcmap(ifst.fst, delta, map_type, power, weight)
    return Fst(newfst)

def compact_symbol_table(syms):
    return pywrapfst.compact_symbol_table(syms)

def compose(ifst1, ifst2, compose_filter='auto', connect=True):
    newfst = pywrapfst.compose(ifst1.fst, ifst2.fst, compose_filter, connect)
    return Fst(newfst)

def convert(ifst, fst_type=None):
    newfst = pywrapfst.convert(ifst.fst, fst_type)
    return Fst(newfst)

def determinize(ifst, delta=1e-6, det_type='functional', nstate=fstlib.NO_STATE_ID, 
                subsequential_label=0, weight=None, incremental_subsequential_label=False):
    newfst = pywrapfst.determinize(ifst.fst, delta, det_type, nstate, subsequential_label, weight, incremental_subsequential_label)
    return Fst(newfst)

def difference(ifst1, ifst2, compose_filter='auto', connect=True):
    newfst = pywrapfst.difference(ifst1.fst, ifst2.fst, compose_filter, connect)
    return Fst(newfst)

def disambiguate(ifst, delta=fstlib.DEF_DELTA, nstate=fstlib.NO_STATE_ID, subsequential_label=0, weight=None):
    newfst = pywrapfst.disambiguate(ifst.fst, delta, nstate, subsequential_label, weight)
    return Fst(newfst)

def divide(lhs, rhs):
    return pywrapfst.divide(lhs, rhs)


def epsnormalize(ifst, eps_norm_output=False):
    newfst = pywrapfst.epsnormalize(ifst.fst, eps_norm_output)
    return Fst(newfst)

def equal(ifst1, ifst2, delta=fstlib.DEF_DELTA):
    return pywrapfst.equal(ifst1.fst, ifst2.fst, delta)

def equivalent(ifst1, ifst2, delta=fstlib.DEF_DELTA):
    return pywrapfst.equivalent(ifst1.fst, ifst2.fst, delta)

def intersect(ifst1, ifst2, compose_filter='auto', connect=True):
    newfst = pywrapfst.intersect(ifst1.fst, ifst2.fst, compose_filter, connect)
    return Fst(newfst)

def isomorphic(ifst1, ifst2, delta=fstlib.DEF_DELTA):
    return pywrapfst.isomorphic(ifst1.fst, ifst2.fst, delta)

def plus(lhs, rhs):
    return pywrapfst.plus(lhs, rhs)

def power(lhs, rhs):
    return pywrapfst.power(lhs, rhs)

def prune(ifst, delta=fstlib.DEF_DELTA, nstate=fstlib.NO_STATE_ID, weight=None):
    newfst = pywrapfst.prune(ifst.fst, delta, nstate, weight)
    return Fst(newfst)

def push(ifst, delta=fstlib.DEF_DELTA, push_weights=True, push_labels=False, 
         remove_common_affix=False, remove_total_weight=False, to_final=False):
    if remove_total_weight:
        logger.warn('remove_total_weight dysfunctional due to bug in pywrapfst. Use destructive method instead.')
    newfst = pywrapfst.push(ifst.fst, delta, push_weights, push_labels, remove_common_affix, remove_total_weight, to_final)
    return Fst(newfst)

def randequivalent(ifst1, ifst2, npath=1, delta=fstlib.DEF_DELTA, seed='auto', select='uniform', max_length=fstlib.MAX_INT32):
    if seed == 'auto':
        seed = _generate_seed()
    else:
        seed = int(seed)
    return pywrapfst.randequivalent(ifst1.fst, ifst2.fst, npath, delta, select, max_length, seed)

def randgen(ifst, npath=1, select='uniform', max_length=fstlib.MAX_INT32, remove_total_weight=False, weighted=False, seed='auto', verbose=False):
    if seed == 'auto':
        seed = _generate_seed()
        if verbose:
            logging.info('Seed = %d' % int(seed))
    else:
        seed = int(seed)
    newfst = pywrapfst.randgen(ifst.fst, npath, select, max_length, remove_total_weight, weighted, seed)
    return Fst(newfst)

def replace(pairs, call_arc_labeling='input',  return_arc_labeling='neither', 
            epsilon_on_replace=False, return_label=0):
    return pywrapfst.replace(pairs, call_arc_labeling, return_arc_labeling, epsilon_on_replace, return_label)

def reverse(ifst, require_superinitial=False):
    newfst = pywrapfst.reverse(ifst.fst, require_superinitial)
    return Fst(newfst)

def shortestdistance(fst, delta=1e-6, nstate=fstlib.NO_STATE_ID, queue_type='auto', reverse=False):
    return pywrapfst.shortestdistance(fst.fst, delta, nstate, queue_type, reverse)

def shortestpath(fst, delta=1e-6, nshortest=1, nstate=fstlib.NO_STATE_ID, queue_type='auto', unique=False, weight=None):
    newfst = pywrapfst.shortestpath(fst.fst, delta, nshortest, nstate, queue_type, unique, weight)
    return Fst(newfst)

def statemap(ifst, map_type):
    newfst = pywrapfst.statemap(ifst.fst, map_type)
    return Fst(newfst)

def synchronize(ifst):
    newfst = pywrapfst.synchronize(ifst.fst)
    return Fst(newfst)

def times(lhs, rhs):
    return pywrapfst.times(lhs, rhs)





import numpy as np
import pandas as pd
import functools
import logging
import fstlib
import fstlib.core
import fstlib.algos

logger = logging.getLogger(__name__)

class FstToolsError(Exception):
    pass

def paths(infst, tape='input'):
    """ returns all strings from a fsa """
    algo = fstlib.algos.PathDepthFirstSearch(infst)
            
    paths=list()
    scores=list()
    for path in algo.get_paths():
        if tape == 'input':
            syms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.input_symbols()])
            seq = [syms[p.ilabel] if syms is not None else str(p.ilabel) for p in path]
        elif tape == 'output':
            syms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.output_symbols()])
            seq = [syms[p.olabel] if syms is not None else str(p.olabel) for p in path]
        elif tape == 'both':
            isyms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.input_symbols()])
            osyms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.output_symbols()])
            seq = [(isyms[p.ilabel] if isyms is not None else str(p.ilabel), 
                    osyms[p.olabel] if osyms is not None else str(p.olabel)) for p in path]
        else:
            raise FstToolsError('Unknown tape parameter %s' % tape)
        score = functools.reduce(fstlib.times, [p.weight for p in path] + [path.finalWeight,])
        paths.append(seq)
        scores.append(float(score))

    result = list(zip(paths, scores))
    return result

def strings(infst, tape='input', delimiter="", to_real=False):
    """ returns all strings from a fsa """
    algo = fstlib.algos.PathDepthFirstSearch(infst)
            
    paths=list()
    scores=list()
    for path in algo.get_paths():
        if tape == 'input':
            syms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.input_symbols()])
            seq = delimiter.join([syms[p.ilabel] if syms is not None else str(p.ilabel) for p in path])
        elif tape == 'output':
            syms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.output_symbols()])
            seq = delimiter.join([syms[p.olabel] if syms is not None else str(p.olabel) for p in path])
        elif tape == 'both':
            isyms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.input_symbols()])
            osyms = dict([(i,s if isinstance(s, str) else str(s, 'utf-8')) for i,s in infst.output_symbols()])
            seq = delimiter.join(["%s:%s" % 
                         (isyms[p.ilabel] if isyms is not None else str(p.ilabel), 
                          osyms[p.olabel] if osyms is not None else str(p.olabel)) for p in path])
        else:
            raise FstToolsError('Unknown tape parameter %s' % tape)
        if infst.arc_type() == fstlib.Semiring.LOG or infst.arc_type() == fstlib.Semiring.TROPICAL:
            score = functools.reduce(fstlib.times, [p.weight for p in path] + [path.finalWeight,])
        else:
            raise FstToolsError('semiring not implemented')
        paths.append(seq)
        scores.append(float(score))
    if len(paths)==0:
        paths.append('[no path]')
        scores.append(np.inf)
    digits = int(np.log10(len(paths)))+1
    result = pd.DataFrame(data=list(zip(paths, scores)), columns=['string','weight'])
    result.sort_values('weight', inplace=True)
    result.index=[("path%%.%dd" % digits) % i for i in range(len(paths))]
    if to_real:
        result.weight = np.exp(-result.weight)

    return result

def mass_intersect_quick(ifsts, prune_nstate = -1, prune_weight = "", prune_level = 0, determinize=True, minimize=True, rmepsilon=True, delta_det=fstlib.DEF_DELTA, delta_min=fstlib.DEF_DELTA, return_filename=False):
    return multicommand(fstlib.core.intersect, ifsts, prune_nstate, prune_weight, prune_level, determinize=determinize, 
                            minimize=minimize,
                          rmepsilon=rmepsilon, sort='olabel', return_filename=return_filename, delta_min=delta_min, 
                          delta_det=delta_det)

def multicommand(func, ifsts, prune_nstate = -1, prune_weight = "", prune_level = 0, determinize=True, minimize=True, rmepsilon=True,
                    sort=None, repeat=1, delta_det=fstlib.DEF_DELTA, delta_min=fstlib.DEF_DELTA, return_filename=False):
    """ applies a function multiple times to multiple fsts or repeatedly to the same.
    Uses Cyril's "log" strategy to speed up computation. """
        
    logger.info("Multicommand %s" % str(func))

    remaining = ifsts[:]
    level = 0
    funcalls = 0
    while len(remaining)>1:
        next_level = remaining[:]
        for i in range(len(remaining)//2): ## build pairs
            ind1 = 2*i
            ind2 = 2*i+1
            ifst1 = remaining[ind1]
            ifst2 = remaining[ind2]
                
            next_level.remove(ifst1)
            next_level.remove(ifst2)

            #logger.info("processing %d and %d." % (ind1, ind2))	
            if sort == 'olabel': ## sort ifst1
                ifst1.arcsort(sort_type=sort)
            elif sort == 'ilabel': ## sort ifst2
                ifst2.arcsort(sort_type=sort)
            else: ## do not sort
                pass
            
            ## do the function
            ofst = func(ifst1, ifst2)
            funcalls += 1

            do_prune = False
            if (prune_weight != "" or prune_nstate >=0) and prune_level <= level:
                do_prune = True

            if not determinize and do_prune: ## if we are determinizing use weighted determinization instead
                ofst = fstlib.core.prune(ofst, weight=prune_weight, nstate=prune_nstate)

            if determinize:
                if rmepsilon:
                    ofst.rmepsilon()
                if do_prune:
                    ofst = fstlib.core.determinize(ofst, delta=delta_det, nstate=prune_nstate, weight=prune_weight)
                else:
                    ofst = fstlib.core.determinize(ofst)
            if minimize:
                ofst.minimize(delta = delta_min)							
            
            next_level.append(ofst)
        remaining = next_level
        level += 1
        
    logger.info("done (%d function calls)!" % funcalls)
        
    return ofst

def _exec_func(func, ifst1, ifst2, delta_min, output, determinize=False, minimize=False, rmepsilon=False, sort='olabel'):
    if sort == 'olabel': ## sort ifst1
        ifst1.arcsort(sort_type=sort)
    elif sort == 'ilabel': ## sort ifst2
        ifst2.arcsort(sort_type=sort)
    else: ## do not sort
        pass
            
    ## do the function
    ofst = func(ifst1, ifst2)

    if determinize:
        if rmepsilon:
            ofst.rmepsilon()
        ofst = fstlib.core.determinize(ofst)
    if minimize:
        ofst.minimize(delta = delta_min)
    output.put(ofst)


def strings_to_count_matrix(list_of_strings, symbols=None):
    """ transforms a list of strings into a count matrix by counting the number of occurences of 
    each symbol in all strings at each position in the list. """
    if symbols is None:
        symbols = list(set(''.join(list_of_strings)))
        symbols.sort()
    n = np.max([len(s) for s in list_of_strings]) ## determine length of longest strings
    m = len(symbols)
    count_matrix = np.zeros((m, n))
    for i in range(n):
        for j in range(len(list_of_strings)):
            s = list_of_strings[j][i]
            k = symbols.index(s)
            count_matrix[k,i] += 1

    return count_matrix, symbols






import logging

import Bio
import Bio.Phylo
import fstlib
import numpy as np

import medicc

logger = logging.getLogger(__name__)

def reconstruct_ancestors(tree, samples_dict, fst, normal_name, prune_weight=0):

    if len(samples_dict) == 2:
        return samples_dict

    fsa_dict = samples_dict.copy()
    tree = Bio.Phylo.BaseTree.copy.deepcopy(tree)

    clade_list = [clade for clade in tree.find_clades(order="preorder") if clade.name != normal_name]
    logger.info("Ancestor reconstruction: Up the tree")
    # up the tree (leaf to root)
    for node in reversed(clade_list): 
        if len(node.clades) != 0:
            children = [item for item in node.clades if item.name != normal_name]
            left_name = children[0].name
            right_name = children[1].name
            logger.debug(f"Clade: {node.name}, left: {left_name}, right: {right_name}")

            ## project
            intersection = intersect_clades_detmin(fsa_dict[left_name], fsa_dict[right_name], fst, 
                                                   prune_weight=prune_weight, detmin_before_intersect=False, detmin_after_intersect=True)
            fsa_dict[node.name] = intersection

    logger.debug("Ancestor reconstruction for root")
    # root node is calculated separately w.r.t. normal node
    root_name = clade_list[0].name 
    sp = fstlib.align(fst, fsa_dict[normal_name], fsa_dict[root_name])
    fsa_dict[root_name] = fstlib.arcmap(sp.copy().project('output'), map_type='rmweight')

    logger.info("Ancestor reconstruction: Down the tree")
    # down the tree (root to leaf)
    for node in clade_list:
        if len(node.clades) != 0:
            children = [q for q in node.clades if len(q.clades) != 0]
            logger.debug(f"Clade: {node.name}, internal children: {children}")
            for child in children:
                sp = fstlib.align(fst, fsa_dict[node.name], fsa_dict[child.name])
                fsa_dict[child.name] = fstlib.arcmap(sp.copy().project('output'), map_type='rmweight')

    # check if ancestors were correctly reconstructed
    sample_lengths = {sample: len(medicc.tools.fsa_to_string(fsa_dict[sample])) for sample, fsa in fsa_dict.items()}
    normal_length = sample_lengths[normal_name]

    if np.any([x != normal_length for x in sample_lengths.values()]):
        raise MEDICCAncestorReconstructionError("Some ancestors could not be reconstructed. These are:\n"
                                                "{}".format('\n'.join([sample for sample, length in sample_lengths.items() if length != normal_length])) + \
                                                "\nCheck whether your normal sample contains segments with copy number zero")

    return fsa_dict


def intersect_clades_detmin(left, right, fst, prune_weight=None, detmin_before_intersect=True, detmin_after_intersect=True):
    L = fstlib.compose(fst, left.arcsort('ilabel')).project('input')
    R = fstlib.compose(fst, right.arcsort('ilabel')).project('input')
    if detmin_before_intersect:
        L = fstlib.determinize(L).minimize()
        R = fstlib.determinize(R).minimize()
    intersection = fstlib.intersect(L.arcsort('olabel'), R)
    # For prune_weight=0, deletes all paths but the shortest one
    if prune_weight is not None:
        pruned = fstlib.prune(intersection, weight=prune_weight)
    else:
        pruned = intersection
    if detmin_after_intersect:
        pruned = fstlib.determinize(pruned).minimize()
    return pruned


class MEDICCAncestorReconstructionError(Exception):
    pass



import logging

import pandas as pd
import numpy as np

import fstlib
import fstlib.cext.ops
from fstlib.cext import pywrapfst

logger = logging.getLogger(__name__)

try:
    from fstlib.cext.ops import *
    logger.info('Using accelerated C operations.')
except ImportError:
    logger.warn('Accelerated C operations not available.')

def align(model, ifst1, ifst2):
    ofst = fstlib.cext.ops.align(model.fst, ifst1.fst, ifst2.fst)
    return fstlib.Fst(ofst)

def encode_determinize_minimize(ifst, delta=1e-6):
    em = fstlib.EncodeMapper(arc_type=ifst.arc_type(), encode_labels=True, encode_weights=False)
    ofst = ifst.copy().encode(em)
    ofst = fstlib.determinize(ofst, delta=delta)
    ofst.minimize()
    ofst.decode(em)
    return ofst

def info(ifst, name=None):
    """ Outputs some info stats """
    info = {}
    info['fst_type'] = ifst.fst_type()
    info['arc_type'] = ifst.arc_type()
    info['nstates'] = ifst.num_states()
    info['nfinalstates'] = np.sum([ifst.is_final(s) for s in ifst.states()])
    info['narcs'] = 0
    info['ninputeps'] = 0
    info['noutputeps'] = 0
    for state in ifst.states():
        info['narcs'] += ifst.num_arcs(state)
        info['ninputeps'] += ifst.num_input_epsilons(state)
        info['noutputeps'] += ifst.num_output_epsilons(state)
    
    df = pd.DataFrame.from_dict(info, orient='index', columns=[name if name is not None else id(ifst)])
    return df

def kernel_score(model, ifst1, ifst2):
    if model.arc_type()=='standard':
        distance = fstlib.cext.ops.kernel_score_std(model.fst, ifst1.fst, ifst2.fst)
    elif model.arc_type()=='log':
        ##distance = fstlib.cext.ops.score_log(model.fst, ifst1.fst, ifst2.fst)
        distance = None
    else:
        distance = None
    return distance

def normalize_alphabet(ifst, inplace=True):
    """Normalizes fst so that outgoing transition weights of the same input symbol sum to 1"""
    
    ## convert to real if necessary
    if inplace:
        ofst = ifst
    else:
        ofst = ifst.copy()
        
    ofst.weight_map(fstlib.algos.map_log_to_real)

    for state in ofst.states():
        arcweights = [(arc.ilabel, float(arc.weight)) for arc in ofst.arcs(state)]
        labelweights = pd.DataFrame.from_records(arcweights, columns=['ilabel','weight']).groupby('ilabel').sum()
        mai = ofst.mutable_arcs(state)
        for arc in mai:
            labelsum = labelweights.weight.loc[arc.ilabel]
            arc.weight = fstlib.Weight(ofst.weight_type(), float(arc.weight) / labelsum)
            mai.set_value(arc)

    ## convert back
    ofst.weight_map(fstlib.algos.map_real_to_log)
    return ofst

def read(source):
    newfst = pywrapfst.Fst.read(source)
    return fstlib.Fst(newfst)

def score(model, ifst1, ifst2):
    if model.arc_type()=='standard':
        distance = fstlib.cext.ops.score_std(model.fst, ifst1.fst, ifst2.fst)
    elif model.arc_type()=='log':
        distance = fstlib.cext.ops.score_log(model.fst, ifst1.fst, ifst2.fst)
    else:
        pass
    return distance

def weight_map(ifst, func):
    newfst = ifst.copy()
    newfst.weight_map(func)
    return newfst



import logging
import random

import Bio
import numpy as np
import pandas as pd
import scipy as sp

from medicc import tools

logger = logging.getLogger(__name__)

def evolve(cnstr, mu=5, plen=0.2, pwgd=0.05, pgain=0.5, maxcn=8, mincn=0, verbose=True, seed=None):
    """ Evolves a copy-number string, e.g. 11111X11111 with the given parameters. """
    nevents = sp.stats.poisson.rvs(mu, size=1)[0]
    cnstr_new = np.array(list(cnstr))
    nloss = ngain = nwgd = 0
    if seed is not None:
        np.random.seed(seed)
    for i in range(nevents):
        if np.random.uniform() <= pwgd: ## WGD
            nwgd += 1
            _mutate(cnstr_new, 'gain', 0, len(cnstr_new), maxcn, mincn, ignore_chr_boundaries=True, verbose=verbose)
        else:
            available_positions = np.flatnonzero((cnstr_new != '0') & (cnstr_new != 'X'))
            if len(available_positions) == 0:
                logger.warning("No available positions to mutate.")
                break
            start = np.random.choice(available_positions)
            length = np.random.geometric(plen)
            end = min(start+length, np.max(available_positions)+1)
            if np.random.uniform() <= pgain: ## gain
                ngain += 1
                _mutate(cnstr_new, 'gain', start, end, maxcn, mincn, verbose=verbose)
            else:
                nloss += 1
                _mutate(cnstr_new, 'loss', start, end, maxcn, mincn, verbose=verbose)
    return ("".join(cnstr_new), nloss, ngain, nwgd)

def _mutate(cnstr, event, start, end, maxcn, mincn, ignore_chr_boundaries=False, verbose=True):
    """ Mutates an individual copy-number string, e.g. 1111X111111. """
    if verbose:
        logger.info("%s event from %d to %d", event, start, end)
    for i in range(start, end):
        if cnstr[i]=='X':
            if ignore_chr_boundaries:
                continue
            else:
                break
        oldcn = tools.hex2int(cnstr[i])
        if event == 'gain':
            if oldcn != 0:
                newcn = oldcn + 1
            else:
                newcn = oldcn
            if newcn > maxcn:
                newcn = maxcn
        elif event == 'loss':
            newcn = oldcn - 1
            if newcn < mincn:
                newcn = mincn
        else:
            logger.error('unknown event')
            newcn = oldcn
        cnstr[i] = tools.int2hex(newcn)

def rcoal(n, tips = None, uniform_branch_length=False):
    """As in rcoal function in R package Ape, function generates a random tree and random branch lengths."""
    x = np.divide(2*np.random.exponential(size = n-1),np.multiply(np.array(list(reversed(range(2,n+1)))), np.array(list(reversed(range(1, n))))))
    if (n == 2):
        edge = pd.DataFrame()
        edge['a'] = [2,2]
        edge['b'] = [0,1]
        edge['name_a'] = pd.Series(['mrca','mrca'])
        if tips and len(tips)==n:
            edge['name_b'] = pd.Series([tips[0],tips[1]])
        else:
            edge['name_b'] = pd.Series(['sp_0001','sp_0002'])
        edge['edge_length'] = np.array([x,x])
        reordered = edge.copy()
    elif (n == 3):
        edge = pd.DataFrame()
        edge['a'] = [3,4,4,3]
        edge['b'] = [4,0,1,2]
        edge['name_a'] = pd.Series(['mrca','internal_1','internal_1','mrca'])
        if tips and len(tips)==n:
            edge['name_b'] = pd.Series(['internal_1',tips[0],tips[1],tips[2]])
        else:
            edge['name_b'] = pd.Series(['internal_1','sp_0001','sp_0002','sp_0003'])
        edge['edge_length'] = np.array([x[1],x[0],x[0],sum(x)])
        reordered = edge.copy()
    else:
        edge = np.zeros((2*n - 2,2), dtype=float)
        edge_length = np.array([0]*(2*n - 2),dtype = float)
        h = np.array([0]*(2*n - 1),dtype = float)
        node_height = np.cumsum(x)
        pool = np.array(list(range(0,n)))# 1:n
        nextnode  = 2*n - 2
        for i in range(0,n-1):
            y = random.sample(list(pool), 2)
            ind = np.array([i*2 , i*2 + 1])
            edge[ind, 1] = y
            edge[ind, 0] = nextnode
            edge_length[ind] = node_height[i] - h[y]
            h[nextnode] = node_height[i]
            pool = np.append(pool[np.isin(pool,y,invert = True)],[nextnode],axis = 0)
            nextnode -= 1
        edge = pd.DataFrame(edge,dtype = object)
        edge.columns = ['a','b']


        reordered = pd.DataFrame()
        root = edge.index[-2]
        parent = [edge.iloc[root][0]] #stack
        child = edge.iloc[root][1]
        reordered = reordered.append(edge.iloc[root])
        edge.drop(edge.index[root], inplace = True)
        while  not edge.empty:
            if (set (edge.a.isin([parent[-1]]))) == {False}:#checks if both occurences took place, and value can be popped out of stack
                del parent[-1]
            if child <= n :
                root = edge.a.isin([parent[-1]]).idxmax()
                child = int(edge.iloc[edge.index==root]['b'])
                reordered = reordered.append(edge.iloc[edge.index==root])
                edge.drop(root, inplace = True)
                if child < n:
                    if (set (edge.a.isin([parent[-1]]))) == {False}:
                        del parent[-1]
            else:
                parent.append(child)
                root =  edge.a.isin([child]).idxmax()
                child = int(edge.iloc[edge.index==root]['b'])
                reordered = reordered.append(edge.iloc[edge.index==root])
                edge.drop(root, inplace = True)
                if child < n:
                    if (set (edge.a.isin([parent[-1]]))) == {False}:
                        del parent[-1]
        reordered['a'] = reordered['a'].astype(int)
        reordered['b'] = reordered['b'].astype(int)
        reordered.loc[:,'edge_length'] = pd.Series(edge_length).reindex(index=reordered.index)
        reordered.replace((reordered['b'].loc[reordered['b']<n].values),list(range(0,n)), inplace = True)
        reordered.edge_length = round(reordered.edge_length,3)
        reordered.index = range(len(reordered))

        if tips and len(tips)==n:
            reordered['name_b'] = [tips[reordered.b[i]] if (reordered.b[i] < n)  else '' for i in reordered.index]
        else:
            reordered['name_b'] = ["sp_{0:04d}".format(reordered.b[i] +1) if (reordered.b[i] < n)  else '' for i in reordered.index]
        reordered.at[reordered.name_b=='', 'name_b'] = ["internal_{0:d}".format(i +1) for i in range(len(reordered[reordered.name_b=='']))]
        internal_names =  reordered[reordered.b>n][['b','name_b']].copy()
        internal_names.set_index(['b'], inplace = True)
        internal_names = internal_names.append(pd.DataFrame(data = ['mrca'], index= [n], columns = ['name_b']))
        reordered['name_a'] = [internal_names.name_b .loc[i]for i in reordered.a]

    reordered = reordered[['a','b','name_a', 'name_b', 'edge_length']]
    if uniform_branch_length:
        reordered['edge_length'] = 1
    return _edgelist_to_tree(reordered)


def _edgelist_to_tree(el, add_normal=True, normal_name='diploid'):
    clades = {name:Bio.Phylo.PhyloXML.Clade(name=name) for name in el[['name_a','name_b']].stack().unique()} 
    if add_normal:
        clades[normal_name] = Bio.Phylo.PhyloXML.Clade(name=normal_name, branch_length=1)
        clades['mrca'].clades.append(clades[normal_name])

    for _, row in el.iterrows():
        parent = clades[row['name_a']]
        child = clades[row['name_b']]
        child.branch_length = row['edge_length']
        parent.clades.append(child)
    
    tree = Bio.Phylo.PhyloXML.Phylogeny(root = clades['mrca'])
    return tree



from dataclasses import dataclass

import fstlib
import numpy as np
import pytest

import medicc


@dataclass
class PairTest:
    seq_in: str
    seq_out: str
    expected_score_asym_wgd: float
    expected_score_asym_no_wgd: float
    expected_score_asym_total_cn: float
    expected_score_sym_wgd: float
    expected_score_sym_no_wgd: float
    expected_score_sym_total_cn: float


PAIR_TESTS = [
    PairTest('11111X1111', '22022X1111', 2, 2, 2, 2, 2, 2),
    PairTest('11111X1111', '22022X2222', 2, 3, 3, 2, 3, 3),
    PairTest('11101X1111', '10111X1111', np.inf, np.inf, np.inf, 2, 2, 2),
    PairTest('33233X1111', '00000X1111', 3, 3, 3, 3, 3, 3),
    PairTest('1111111111X1111111111', '2212222222X2222222222', 2, 3, 3, 2, 3, 3),
    PairTest('2222222222X2222222222', '3323333333X3333323333', 3, 4, 4, 3, 4, 4),
    PairTest('1111111X11X11X1111X1111', '3322112X22X23X2222X2200', 5, 9, 8, 5, 9, 8),
    PairTest('1111111111X111X111', '3332222221X333X333', 4, 6, 3, 4, 6, 3),
    PairTest('22222X222X222', '44444X444X444', 2, 6, 1, 2, 6, 1),
]


def _run_pair_test(pair_test: PairTest, is_wgd: bool, is_sym: bool, is_total_cn: bool) -> bool:
    """Runs individual pair test"""
    # maxcn = 8  # alphabet = 012345678; maxcn losses, maxcn-1 gains
    # sep = "X"
    # symbol_table = medicc.create_symbol_table(maxcn, sep)
    # fst = medicc.create_copynumber_fst(symbol_table, sep, enable_wgd=is_wgd)
    fst = medicc.io.read_fst(no_wgd=(not is_wgd), total_copy_numbers=is_total_cn, wgd_x2=False)

    td = fstlib.factory.from_string(pair_test.seq_in, isymbols=fst.input_symbols(), osymbols=fst.output_symbols(), arc_type=fst.arc_type())
    tg = fstlib.factory.from_string(pair_test.seq_out, isymbols=fst.input_symbols(), osymbols=fst.output_symbols(), arc_type=fst.arc_type())
    if is_sym:
        test_score = float(fstlib.kernel_score(fst, td, tg))
        if is_total_cn:
            expected_score = pair_test.expected_score_sym_total_cn
        elif is_wgd:
            expected_score = pair_test.expected_score_sym_wgd
        else:
            expected_score = pair_test.expected_score_sym_no_wgd
    else:
        test_score = float(fstlib.score(fst, td, tg))
        if is_total_cn:
            expected_score = pair_test.expected_score_asym_total_cn
        elif is_wgd:
            expected_score = pair_test.expected_score_asym_wgd
        else:
            expected_score = pair_test.expected_score_asym_no_wgd
    return test_score, expected_score


@pytest.mark.parametrize("pair", PAIR_TESTS)
def test_fstlib_sym_with_wgd(pair: PairTest):

    test_score, expected_score = _run_pair_test(pair, is_wgd=True, is_sym=True, is_total_cn=False)
    assert test_score == expected_score

@pytest.mark.parametrize("pair", PAIR_TESTS)
def test_fstlib_sym_without_wgd(pair: PairTest):

    test_score, expected_score = _run_pair_test(pair, is_wgd=False, is_sym=True, is_total_cn=False)
    assert test_score == expected_score

@pytest.mark.parametrize("pair", PAIR_TESTS)
def test_fstlib_asym_with_wgd(pair: PairTest):

    test_score, expected_score = _run_pair_test(pair, is_wgd=True, is_sym=False, is_total_cn=False)
    assert test_score == expected_score

@pytest.mark.parametrize("pair", PAIR_TESTS)
def test_fstlib_asym_without_wgd(pair: PairTest):

    test_score, expected_score = _run_pair_test(pair, is_wgd=False, is_sym=False, is_total_cn=False)
    assert test_score == expected_score

@pytest.mark.parametrize("pair", PAIR_TESTS)
def test_fstlib_asym_total_cn(pair: PairTest):

    test_score, expected_score = _run_pair_test(pair, is_wgd=True, is_sym=False, is_total_cn=True)
    assert test_score == expected_score

@pytest.mark.parametrize("pair", PAIR_TESTS)
def test_fstlib_sym_total_cn(pair: PairTest):

    test_score, expected_score = _run_pair_test(pair, is_wgd=True, is_sym=True, is_total_cn=True)
    assert test_score == expected_score
